# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Runtime

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


class ScriptId(str):
    """
    Unique script identifier.
    """

    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> ScriptId:
        return cls(json)

    def __repr__(self):
        return "ScriptId({})".format(super().__repr__())


@dataclass
class SerializationOptions:
    """
    Represents options for serialization. Overrides ``generatePreview`` and ``returnByValue``.
    """

    serialization: str

    #: Deep serialization depth. Default is full depth. Respected only in ``deep`` serialization mode.
    max_depth: typing.Optional[int] = None

    #: Embedder-specific parameters. For example if connected to V8 in Chrome these control DOM
    #: serialization via ``maxNodeDepth: integer`` and ``includeShadowTree: "none" `` "open" `` "all"``.
    #: Values can be only of type string or integer.
    additional_parameters: typing.Optional[dict] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["serialization"] = self.serialization
        if self.max_depth is not None:
            json["maxDepth"] = self.max_depth
        if self.additional_parameters is not None:
            json["additionalParameters"] = self.additional_parameters
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SerializationOptions:
        return cls(
            serialization=str(json["serialization"]),
            max_depth=int(json["maxDepth"])
            if json.get("maxDepth", None) is not None
            else None,
            additional_parameters=dict(json["additionalParameters"])
            if json.get("additionalParameters", None) is not None
            else None,
        )


@dataclass
class DeepSerializedValue:
    """
    Represents deep serialized value.
    """

    type_: str

    value: typing.Optional[typing.Any] = None

    object_id: typing.Optional[str] = None

    #: Set if value reference met more then once during serialization. In such
    #: case, value is provided only to one of the serialized values. Unique
    #: per value in the scope of one CDP call.
    weak_local_object_reference: typing.Optional[int] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["type"] = self.type_
        if self.value is not None:
            json["value"] = self.value
        if self.object_id is not None:
            json["objectId"] = self.object_id
        if self.weak_local_object_reference is not None:
            json["weakLocalObjectReference"] = self.weak_local_object_reference
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DeepSerializedValue:
        return cls(
            type_=str(json["type"]),
            value=json["value"] if json.get("value", None) is not None else None,
            object_id=str(json["objectId"])
            if json.get("objectId", None) is not None
            else None,
            weak_local_object_reference=int(json["weakLocalObjectReference"])
            if json.get("weakLocalObjectReference", None) is not None
            else None,
        )


class RemoteObjectId(str):
    """
    Unique object identifier.
    """

    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> RemoteObjectId:
        return cls(json)

    def __repr__(self):
        return "RemoteObjectId({})".format(super().__repr__())


class UnserializableValue(str):
    """
    Primitive value which cannot be JSON-stringified. Includes values ``-0``, ``NaN``, ``Infinity``,
    ``-Infinity``, and bigint literals.
    """

    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> UnserializableValue:
        return cls(json)

    def __repr__(self):
        return "UnserializableValue({})".format(super().__repr__())


@dataclass
class RemoteObject:
    """
    Mirror object referencing original JavaScript object.
    """

    #: Object type.
    type_: str

    #: Object subtype hint. Specified for ``object`` type values only.
    #: NOTE: If you change anything here, make sure to also update
    #: ``subtype`` in ``ObjectPreview`` and ``PropertyPreview`` below.
    subtype: typing.Optional[str] = None

    #: Object class (constructor) name. Specified for ``object`` type values only.
    class_name: typing.Optional[str] = None

    #: Remote object value in case of primitive values or JSON values (if it was requested).
    value: typing.Optional[typing.Any] = None

    #: Primitive value which can not be JSON-stringified does not have ``value``, but gets this
    #: property.
    unserializable_value: typing.Optional[UnserializableValue] = None

    #: String representation of the object.
    description: typing.Optional[str] = None

    #: Deep serialized value.
    deep_serialized_value: typing.Optional[DeepSerializedValue] = None

    #: Unique object identifier (for non-primitive values).
    object_id: typing.Optional[RemoteObjectId] = None

    #: Preview containing abbreviated property values. Specified for ``object`` type values only.
    preview: typing.Optional[ObjectPreview] = None

    custom_preview: typing.Optional[CustomPreview] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["type"] = self.type_
        if self.subtype is not None:
            json["subtype"] = self.subtype
        if self.class_name is not None:
            json["className"] = self.class_name
        if self.value is not None:
            json["value"] = self.value
        if self.unserializable_value is not None:
            json["unserializableValue"] = self.unserializable_value.to_json()
        if self.description is not None:
            json["description"] = self.description
        if self.deep_serialized_value is not None:
            json["deepSerializedValue"] = self.deep_serialized_value.to_json()
        if self.object_id is not None:
            json["objectId"] = self.object_id.to_json()
        if self.preview is not None:
            json["preview"] = self.preview.to_json()
        if self.custom_preview is not None:
            json["customPreview"] = self.custom_preview.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> RemoteObject:
        return cls(
            type_=str(json["type"]),
            subtype=str(json["subtype"])
            if json.get("subtype", None) is not None
            else None,
            class_name=str(json["className"])
            if json.get("className", None) is not None
            else None,
            value=json["value"] if json.get("value", None) is not None else None,
            unserializable_value=UnserializableValue.from_json(
                json["unserializableValue"]
            )
            if json.get("unserializableValue", None) is not None
            else None,
            description=str(json["description"])
            if json.get("description", None) is not None
            else None,
            deep_serialized_value=DeepSerializedValue.from_json(
                json["deepSerializedValue"]
            )
            if json.get("deepSerializedValue", None) is not None
            else None,
            object_id=RemoteObjectId.from_json(json["objectId"])
            if json.get("objectId", None) is not None
            else None,
            preview=ObjectPreview.from_json(json["preview"])
            if json.get("preview", None) is not None
            else None,
            custom_preview=CustomPreview.from_json(json["customPreview"])
            if json.get("customPreview", None) is not None
            else None,
        )


@dataclass
class CustomPreview:
    #: The JSON-stringified result of formatter.header(object, config) call.
    #: It contains json ML array that represents RemoteObject.
    header: str

    #: If formatter returns true as a result of formatter.hasBody call then bodyGetterId will
    #: contain RemoteObjectId for the function that returns result of formatter.body(object, config) call.
    #: The result value is json ML array.
    body_getter_id: typing.Optional[RemoteObjectId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["header"] = self.header
        if self.body_getter_id is not None:
            json["bodyGetterId"] = self.body_getter_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CustomPreview:
        return cls(
            header=str(json["header"]),
            body_getter_id=RemoteObjectId.from_json(json["bodyGetterId"])
            if json.get("bodyGetterId", None) is not None
            else None,
        )


@dataclass
class ObjectPreview:
    """
    Object containing abbreviated remote object value.
    """

    #: Object type.
    type_: str

    #: True iff some of the properties or entries of the original object did not fit.
    overflow: bool

    #: List of the properties.
    properties: typing.List[PropertyPreview]

    #: Object subtype hint. Specified for ``object`` type values only.
    subtype: typing.Optional[str] = None

    #: String representation of the object.
    description: typing.Optional[str] = None

    #: List of the entries. Specified for ``map`` and ``set`` subtype values only.
    entries: typing.Optional[typing.List[EntryPreview]] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["type"] = self.type_
        json["overflow"] = self.overflow
        json["properties"] = [i.to_json() for i in self.properties]
        if self.subtype is not None:
            json["subtype"] = self.subtype
        if self.description is not None:
            json["description"] = self.description
        if self.entries is not None:
            json["entries"] = [i.to_json() for i in self.entries]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ObjectPreview:
        return cls(
            type_=str(json["type"]),
            overflow=bool(json["overflow"]),
            properties=[PropertyPreview.from_json(i) for i in json["properties"]],
            subtype=str(json["subtype"])
            if json.get("subtype", None) is not None
            else None,
            description=str(json["description"])
            if json.get("description", None) is not None
            else None,
            entries=[EntryPreview.from_json(i) for i in json["entries"]]
            if json.get("entries", None) is not None
            else None,
        )


@dataclass
class PropertyPreview:
    #: Property name.
    name: str

    #: Object type. Accessor means that the property itself is an accessor property.
    type_: str

    #: User-friendly property value string.
    value: typing.Optional[str] = None

    #: Nested value preview.
    value_preview: typing.Optional[ObjectPreview] = None

    #: Object subtype hint. Specified for ``object`` type values only.
    subtype: typing.Optional[str] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        json["type"] = self.type_
        if self.value is not None:
            json["value"] = self.value
        if self.value_preview is not None:
            json["valuePreview"] = self.value_preview.to_json()
        if self.subtype is not None:
            json["subtype"] = self.subtype
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PropertyPreview:
        return cls(
            name=str(json["name"]),
            type_=str(json["type"]),
            value=str(json["value"]) if json.get("value", None) is not None else None,
            value_preview=ObjectPreview.from_json(json["valuePreview"])
            if json.get("valuePreview", None) is not None
            else None,
            subtype=str(json["subtype"])
            if json.get("subtype", None) is not None
            else None,
        )


@dataclass
class EntryPreview:
    #: Preview of the value.
    value: ObjectPreview

    #: Preview of the key. Specified for map-like collection entries.
    key: typing.Optional[ObjectPreview] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["value"] = self.value.to_json()
        if self.key is not None:
            json["key"] = self.key.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> EntryPreview:
        return cls(
            value=ObjectPreview.from_json(json["value"]),
            key=ObjectPreview.from_json(json["key"])
            if json.get("key", None) is not None
            else None,
        )


@dataclass
class PropertyDescriptor:
    """
    Object property descriptor.
    """

    #: Property name or symbol description.
    name: str

    #: True if the type of this property descriptor may be changed and if the property may be
    #: deleted from the corresponding object.
    configurable: bool

    #: True if this property shows up during enumeration of the properties on the corresponding
    #: object.
    enumerable: bool

    #: The value associated with the property.
    value: typing.Optional[RemoteObject] = None

    #: True if the value associated with the property may be changed (data descriptors only).
    writable: typing.Optional[bool] = None

    #: A function which serves as a getter for the property, or ``undefined`` if there is no getter
    #: (accessor descriptors only).
    get: typing.Optional[RemoteObject] = None

    #: A function which serves as a setter for the property, or ``undefined`` if there is no setter
    #: (accessor descriptors only).
    set_: typing.Optional[RemoteObject] = None

    #: True if the result was thrown during the evaluation.
    was_thrown: typing.Optional[bool] = None

    #: True if the property is owned for the object.
    is_own: typing.Optional[bool] = None

    #: Property symbol object, if the property is of the ``symbol`` type.
    symbol: typing.Optional[RemoteObject] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        json["configurable"] = self.configurable
        json["enumerable"] = self.enumerable
        if self.value is not None:
            json["value"] = self.value.to_json()
        if self.writable is not None:
            json["writable"] = self.writable
        if self.get is not None:
            json["get"] = self.get.to_json()
        if self.set_ is not None:
            json["set"] = self.set_.to_json()
        if self.was_thrown is not None:
            json["wasThrown"] = self.was_thrown
        if self.is_own is not None:
            json["isOwn"] = self.is_own
        if self.symbol is not None:
            json["symbol"] = self.symbol.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PropertyDescriptor:
        return cls(
            name=str(json["name"]),
            configurable=bool(json["configurable"]),
            enumerable=bool(json["enumerable"]),
            value=RemoteObject.from_json(json["value"])
            if json.get("value", None) is not None
            else None,
            writable=bool(json["writable"])
            if json.get("writable", None) is not None
            else None,
            get=RemoteObject.from_json(json["get"])
            if json.get("get", None) is not None
            else None,
            set_=RemoteObject.from_json(json["set"])
            if json.get("set", None) is not None
            else None,
            was_thrown=bool(json["wasThrown"])
            if json.get("wasThrown", None) is not None
            else None,
            is_own=bool(json["isOwn"]) if json.get("isOwn", None) is not None else None,
            symbol=RemoteObject.from_json(json["symbol"])
            if json.get("symbol", None) is not None
            else None,
        )


@dataclass
class InternalPropertyDescriptor:
    """
    Object internal property descriptor. This property isn't normally visible in JavaScript code.
    """

    #: Conventional property name.
    name: str

    #: The value associated with the property.
    value: typing.Optional[RemoteObject] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        if self.value is not None:
            json["value"] = self.value.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> InternalPropertyDescriptor:
        return cls(
            name=str(json["name"]),
            value=RemoteObject.from_json(json["value"])
            if json.get("value", None) is not None
            else None,
        )


@dataclass
class PrivatePropertyDescriptor:
    """
    Object private field descriptor.
    """

    #: Private property name.
    name: str

    #: The value associated with the private property.
    value: typing.Optional[RemoteObject] = None

    #: A function which serves as a getter for the private property,
    #: or ``undefined`` if there is no getter (accessor descriptors only).
    get: typing.Optional[RemoteObject] = None

    #: A function which serves as a setter for the private property,
    #: or ``undefined`` if there is no setter (accessor descriptors only).
    set_: typing.Optional[RemoteObject] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        if self.value is not None:
            json["value"] = self.value.to_json()
        if self.get is not None:
            json["get"] = self.get.to_json()
        if self.set_ is not None:
            json["set"] = self.set_.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PrivatePropertyDescriptor:
        return cls(
            name=str(json["name"]),
            value=RemoteObject.from_json(json["value"])
            if json.get("value", None) is not None
            else None,
            get=RemoteObject.from_json(json["get"])
            if json.get("get", None) is not None
            else None,
            set_=RemoteObject.from_json(json["set"])
            if json.get("set", None) is not None
            else None,
        )


@dataclass
class CallArgument:
    """
    Represents function call argument. Either remote object id ``objectId``, primitive ``value``,
    unserializable primitive value or neither of (for undefined) them should be specified.
    """

    #: Primitive value or serializable javascript object.
    value: typing.Optional[typing.Any] = None

    #: Primitive value which can not be JSON-stringified.
    unserializable_value: typing.Optional[UnserializableValue] = None

    #: Remote object handle.
    object_id: typing.Optional[RemoteObjectId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        if self.value is not None:
            json["value"] = self.value
        if self.unserializable_value is not None:
            json["unserializableValue"] = self.unserializable_value.to_json()
        if self.object_id is not None:
            json["objectId"] = self.object_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CallArgument:
        return cls(
            value=json["value"] if json.get("value", None) is not None else None,
            unserializable_value=UnserializableValue.from_json(
                json["unserializableValue"]
            )
            if json.get("unserializableValue", None) is not None
            else None,
            object_id=RemoteObjectId.from_json(json["objectId"])
            if json.get("objectId", None) is not None
            else None,
        )


class ExecutionContextId(int):
    """
    Id of an execution context.
    """

    def to_json(self) -> int:
        return self

    @classmethod
    def from_json(cls, json: int) -> ExecutionContextId:
        return cls(json)

    def __repr__(self):
        return "ExecutionContextId({})".format(super().__repr__())


@dataclass
class ExecutionContextDescription:
    """
    Description of an isolated world.
    """

    #: Unique id of the execution context. It can be used to specify in which execution context
    #: script evaluation should be performed.
    id_: ExecutionContextId

    #: Execution context origin.
    origin: str

    #: Human readable name describing given context.
    name: str

    #: A system-unique execution context identifier. Unlike the id, this is unique across
    #: multiple processes, so can be reliably used to identify specific context while backend
    #: performs a cross-process navigation.
    unique_id: str

    #: Embedder-specific auxiliary data likely matching {isDefault: boolean, type: 'default'``'isolated'``'worker', frameId: string}
    aux_data: typing.Optional[dict] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["id"] = self.id_.to_json()
        json["origin"] = self.origin
        json["name"] = self.name
        json["uniqueId"] = self.unique_id
        if self.aux_data is not None:
            json["auxData"] = self.aux_data
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExecutionContextDescription:
        return cls(
            id_=ExecutionContextId.from_json(json["id"]),
            origin=str(json["origin"]),
            name=str(json["name"]),
            unique_id=str(json["uniqueId"]),
            aux_data=dict(json["auxData"])
            if json.get("auxData", None) is not None
            else None,
        )


@dataclass
class ExceptionDetails:
    """
    Detailed information about exception (or error) that was thrown during script compilation or
    execution.
    """

    #: Exception id.
    exception_id: int

    #: Exception text, which should be used together with exception object when available.
    text: str

    #: Line number of the exception location (0-based).
    line_number: int

    #: Column number of the exception location (0-based).
    column_number: int

    #: Script ID of the exception location.
    script_id: typing.Optional[ScriptId] = None

    #: URL of the exception location, to be used when the script was not reported.
    url: typing.Optional[str] = None

    #: JavaScript stack trace if available.
    stack_trace: typing.Optional[StackTrace] = None

    #: Exception object if available.
    exception: typing.Optional[RemoteObject] = None

    #: Identifier of the context where exception happened.
    execution_context_id: typing.Optional[ExecutionContextId] = None

    #: Dictionary with entries of meta data that the client associated
    #: with this exception, such as information about associated network
    #: requests, etc.
    exception_meta_data: typing.Optional[dict] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["exceptionId"] = self.exception_id
        json["text"] = self.text
        json["lineNumber"] = self.line_number
        json["columnNumber"] = self.column_number
        if self.script_id is not None:
            json["scriptId"] = self.script_id.to_json()
        if self.url is not None:
            json["url"] = self.url
        if self.stack_trace is not None:
            json["stackTrace"] = self.stack_trace.to_json()
        if self.exception is not None:
            json["exception"] = self.exception.to_json()
        if self.execution_context_id is not None:
            json["executionContextId"] = self.execution_context_id.to_json()
        if self.exception_meta_data is not None:
            json["exceptionMetaData"] = self.exception_meta_data
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExceptionDetails:
        return cls(
            exception_id=int(json["exceptionId"]),
            text=str(json["text"]),
            line_number=int(json["lineNumber"]),
            column_number=int(json["columnNumber"]),
            script_id=ScriptId.from_json(json["scriptId"])
            if json.get("scriptId", None) is not None
            else None,
            url=str(json["url"]) if json.get("url", None) is not None else None,
            stack_trace=StackTrace.from_json(json["stackTrace"])
            if json.get("stackTrace", None) is not None
            else None,
            exception=RemoteObject.from_json(json["exception"])
            if json.get("exception", None) is not None
            else None,
            execution_context_id=ExecutionContextId.from_json(
                json["executionContextId"]
            )
            if json.get("executionContextId", None) is not None
            else None,
            exception_meta_data=dict(json["exceptionMetaData"])
            if json.get("exceptionMetaData", None) is not None
            else None,
        )


class Timestamp(float):
    """
    Number of milliseconds since epoch.
    """

    def to_json(self) -> float:
        return self

    @classmethod
    def from_json(cls, json: float) -> Timestamp:
        return cls(json)

    def __repr__(self):
        return "Timestamp({})".format(super().__repr__())


class TimeDelta(float):
    """
    Number of milliseconds.
    """

    def to_json(self) -> float:
        return self

    @classmethod
    def from_json(cls, json: float) -> TimeDelta:
        return cls(json)

    def __repr__(self):
        return "TimeDelta({})".format(super().__repr__())


@dataclass
class CallFrame:
    """
    Stack entry for runtime errors and assertions.
    """

    #: JavaScript function name.
    function_name: str

    #: JavaScript script id.
    script_id: ScriptId

    #: JavaScript script name or url.
    url: str

    #: JavaScript script line number (0-based).
    line_number: int

    #: JavaScript script column number (0-based).
    column_number: int

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["functionName"] = self.function_name
        json["scriptId"] = self.script_id.to_json()
        json["url"] = self.url
        json["lineNumber"] = self.line_number
        json["columnNumber"] = self.column_number
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CallFrame:
        return cls(
            function_name=str(json["functionName"]),
            script_id=ScriptId.from_json(json["scriptId"]),
            url=str(json["url"]),
            line_number=int(json["lineNumber"]),
            column_number=int(json["columnNumber"]),
        )


@dataclass
class StackTrace:
    """
    Call frames for assertions or error messages.
    """

    #: JavaScript function name.
    call_frames: typing.List[CallFrame]

    #: String label of this stack trace. For async traces this may be a name of the function that
    #: initiated the async call.
    description: typing.Optional[str] = None

    #: Asynchronous JavaScript stack trace that preceded this stack, if available.
    parent: typing.Optional[StackTrace] = None

    #: Asynchronous JavaScript stack trace that preceded this stack, if available.
    parent_id: typing.Optional[StackTraceId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["callFrames"] = [i.to_json() for i in self.call_frames]
        if self.description is not None:
            json["description"] = self.description
        if self.parent is not None:
            json["parent"] = self.parent.to_json()
        if self.parent_id is not None:
            json["parentId"] = self.parent_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> StackTrace:
        return cls(
            call_frames=[CallFrame.from_json(i) for i in json["callFrames"]],
            description=str(json["description"])
            if json.get("description", None) is not None
            else None,
            parent=StackTrace.from_json(json["parent"])
            if json.get("parent", None) is not None
            else None,
            parent_id=StackTraceId.from_json(json["parentId"])
            if json.get("parentId", None) is not None
            else None,
        )


class UniqueDebuggerId(str):
    """
    Unique identifier of current debugger.
    """

    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> UniqueDebuggerId:
        return cls(json)

    def __repr__(self):
        return "UniqueDebuggerId({})".format(super().__repr__())


@dataclass
class StackTraceId:
    """
    If ``debuggerId`` is set stack trace comes from another debugger and can be resolved there. This
    allows to track cross-debugger calls. See ``Runtime.StackTrace`` and ``Debugger.paused`` for usages.
    """

    id_: str

    debugger_id: typing.Optional[UniqueDebuggerId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["id"] = self.id_
        if self.debugger_id is not None:
            json["debuggerId"] = self.debugger_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> StackTraceId:
        return cls(
            id_=str(json["id"]),
            debugger_id=UniqueDebuggerId.from_json(json["debuggerId"])
            if json.get("debuggerId", None) is not None
            else None,
        )


def await_promise(
    promise_object_id: RemoteObjectId,
    return_by_value: typing.Optional[bool] = None,
    generate_preview: typing.Optional[bool] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[RemoteObject, typing.Optional[ExceptionDetails]],
]:
    """
    Add handler to promise with given promise object id.

    :param promise_object_id: Identifier of the promise.
    :param return_by_value: *(Optional)* Whether the result is expected to be a JSON object that should be sent by value.
    :param generate_preview: *(Optional)* Whether preview should be generated for the result.
    :returns: A tuple with the following items:

        0. **result** - Promise result. Will contain rejected value if promise was rejected.
        1. **exceptionDetails** - *(Optional)* Exception details if stack strace is available.
    """
    params: T_JSON_DICT = dict()
    params["promiseObjectId"] = promise_object_id.to_json()
    if return_by_value is not None:
        params["returnByValue"] = return_by_value
    if generate_preview is not None:
        params["generatePreview"] = generate_preview
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.awaitPromise",
        "params": params,
    }
    json = yield cmd_dict
    return (
        RemoteObject.from_json(json["result"]),
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def call_function_on(
    function_declaration: str,
    object_id: typing.Optional[RemoteObjectId] = None,
    arguments: typing.Optional[typing.List[CallArgument]] = None,
    silent: typing.Optional[bool] = None,
    return_by_value: typing.Optional[bool] = None,
    generate_preview: typing.Optional[bool] = None,
    user_gesture: typing.Optional[bool] = None,
    await_promise: typing.Optional[bool] = None,
    execution_context_id: typing.Optional[ExecutionContextId] = None,
    object_group: typing.Optional[str] = None,
    throw_on_side_effect: typing.Optional[bool] = None,
    unique_context_id: typing.Optional[str] = None,
    serialization_options: typing.Optional[SerializationOptions] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[RemoteObject, typing.Optional[ExceptionDetails]],
]:
    """
    Calls function with given declaration on the given object. Object group of the result is
    inherited from the target object.

    :param function_declaration: Declaration of the function to call.
    :param object_id: *(Optional)* Identifier of the object to call function on. Either objectId or executionContextId should be specified.
    :param arguments: *(Optional)* Call arguments. All call arguments must belong to the same JavaScript world as the target object.
    :param silent: *(Optional)* In silent mode exceptions thrown during evaluation are not reported and do not pause execution. Overrides ```setPauseOnException```` state.
    :param return_by_value: *(Optional)* Whether the result is expected to be a JSON object which should be sent by value. Can be overriden by ````serializationOptions````.
    :param generate_preview: **(EXPERIMENTAL)** *(Optional)* Whether preview should be generated for the result.
    :param user_gesture: *(Optional)* Whether execution should be treated as initiated by user in the UI.
    :param await_promise: *(Optional)* Whether execution should ````await```` for resulting value and return once awaited promise is resolved.
    :param execution_context_id: *(Optional)* Specifies execution context which global object will be used to call function on. Either executionContextId or objectId should be specified.
    :param object_group: *(Optional)* Symbolic group name that can be used to release multiple objects. If objectGroup is not specified and objectId is, objectGroup will be inherited from object.
    :param throw_on_side_effect: **(EXPERIMENTAL)** *(Optional)* Whether to throw an exception if side effect cannot be ruled out during evaluation.
    :param unique_context_id: **(EXPERIMENTAL)** *(Optional)* An alternative way to specify the execution context to call function on. Compared to contextId that may be reused across processes, this is guaranteed to be system-unique, so it can be used to prevent accidental function call in context different than intended (e.g. as a result of navigation across process boundaries). This is mutually exclusive with ````executionContextId````.
    :param serialization_options: **(EXPERIMENTAL)** *(Optional)* Specifies the result serialization. If provided, overrides ````generatePreview```` and ````returnByValue```.
    :returns: A tuple with the following items:

        0. **result** - Call result.
        1. **exceptionDetails** - *(Optional)* Exception details.
    """
    params: T_JSON_DICT = dict()
    params["functionDeclaration"] = function_declaration
    if object_id is not None:
        params["objectId"] = object_id.to_json()
    if arguments is not None:
        params["arguments"] = [i.to_json() for i in arguments]
    if silent is not None:
        params["silent"] = silent
    if return_by_value is not None:
        params["returnByValue"] = return_by_value
    if generate_preview is not None:
        params["generatePreview"] = generate_preview
    if user_gesture is not None:
        params["userGesture"] = user_gesture
    if await_promise is not None:
        params["awaitPromise"] = await_promise
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    if object_group is not None:
        params["objectGroup"] = object_group
    if throw_on_side_effect is not None:
        params["throwOnSideEffect"] = throw_on_side_effect
    if unique_context_id is not None:
        params["uniqueContextId"] = unique_context_id
    if serialization_options is not None:
        params["serializationOptions"] = serialization_options.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.callFunctionOn",
        "params": params,
    }
    json = yield cmd_dict
    return (
        RemoteObject.from_json(json["result"]),
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def compile_script(
    expression: str,
    source_url: str,
    persist_script: bool,
    execution_context_id: typing.Optional[ExecutionContextId] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[typing.Optional[ScriptId], typing.Optional[ExceptionDetails]],
]:
    """
    Compiles expression.

    :param expression: Expression to compile.
    :param source_url: Source url to be set for the script.
    :param persist_script: Specifies whether the compiled script should be persisted.
    :param execution_context_id: *(Optional)* Specifies in which execution context to perform script run. If the parameter is omitted the evaluation will be performed in the context of the inspected page.
    :returns: A tuple with the following items:

        0. **scriptId** - *(Optional)* Id of the script.
        1. **exceptionDetails** - *(Optional)* Exception details.
    """
    params: T_JSON_DICT = dict()
    params["expression"] = expression
    params["sourceURL"] = source_url
    params["persistScript"] = persist_script
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.compileScript",
        "params": params,
    }
    json = yield cmd_dict
    return (
        ScriptId.from_json(json["scriptId"])
        if json.get("scriptId", None) is not None
        else None,
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disables reporting of execution contexts creation.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.disable",
    }
    json = yield cmd_dict


def discard_console_entries() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Discards collected exceptions and console API calls.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.discardConsoleEntries",
    }
    json = yield cmd_dict


def enable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables reporting of execution contexts creation by means of ``executionContextCreated`` event.
    When the reporting gets enabled the event will be sent immediately for each existing execution
    context.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.enable",
    }
    json = yield cmd_dict


def evaluate(
    expression: str,
    object_group: typing.Optional[str] = None,
    include_command_line_api: typing.Optional[bool] = None,
    silent: typing.Optional[bool] = None,
    context_id: typing.Optional[ExecutionContextId] = None,
    return_by_value: typing.Optional[bool] = None,
    generate_preview: typing.Optional[bool] = None,
    user_gesture: typing.Optional[bool] = None,
    await_promise: typing.Optional[bool] = None,
    throw_on_side_effect: typing.Optional[bool] = None,
    timeout: typing.Optional[TimeDelta] = None,
    disable_breaks: typing.Optional[bool] = None,
    repl_mode: typing.Optional[bool] = None,
    allow_unsafe_eval_blocked_by_csp: typing.Optional[bool] = None,
    unique_context_id: typing.Optional[str] = None,
    serialization_options: typing.Optional[SerializationOptions] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[RemoteObject, typing.Optional[ExceptionDetails]],
]:
    """
    Evaluates expression on global object.

    :param expression: Expression to evaluate.
    :param object_group: *(Optional)* Symbolic group name that can be used to release multiple objects.
    :param include_command_line_api: *(Optional)* Determines whether Command Line API should be available during the evaluation.
    :param silent: *(Optional)* In silent mode exceptions thrown during evaluation are not reported and do not pause execution. Overrides ```setPauseOnException```` state.
    :param context_id: *(Optional)* Specifies in which execution context to perform evaluation. If the parameter is omitted the evaluation will be performed in the context of the inspected page. This is mutually exclusive with ````uniqueContextId````, which offers an alternative way to identify the execution context that is more reliable in a multi-process environment.
    :param return_by_value: *(Optional)* Whether the result is expected to be a JSON object that should be sent by value.
    :param generate_preview: **(EXPERIMENTAL)** *(Optional)* Whether preview should be generated for the result.
    :param user_gesture: *(Optional)* Whether execution should be treated as initiated by user in the UI.
    :param await_promise: *(Optional)* Whether execution should ````await```` for resulting value and return once awaited promise is resolved.
    :param throw_on_side_effect: **(EXPERIMENTAL)** *(Optional)* Whether to throw an exception if side effect cannot be ruled out during evaluation. This implies ````disableBreaks```` below.
    :param timeout: **(EXPERIMENTAL)** *(Optional)* Terminate execution after timing out (number of milliseconds).
    :param disable_breaks: **(EXPERIMENTAL)** *(Optional)* Disable breakpoints during execution.
    :param repl_mode: **(EXPERIMENTAL)** *(Optional)* Setting this flag to true enables ````let```` re-declaration and top-level ````await````. Note that ````let```` variables can only be re-declared if they originate from ````replMode```` themselves.
    :param allow_unsafe_eval_blocked_by_csp: **(EXPERIMENTAL)** *(Optional)* The Content Security Policy (CSP) for the target might block 'unsafe-eval' which includes eval(), Function(), setTimeout() and setInterval() when called with non-callable arguments. This flag bypasses CSP for this evaluation and allows unsafe-eval. Defaults to true.
    :param unique_context_id: **(EXPERIMENTAL)** *(Optional)* An alternative way to specify the execution context to evaluate in. Compared to contextId that may be reused across processes, this is guaranteed to be system-unique, so it can be used to prevent accidental evaluation of the expression in context different than intended (e.g. as a result of navigation across process boundaries). This is mutually exclusive with ````contextId````.
    :param serialization_options: **(EXPERIMENTAL)** *(Optional)* Specifies the result serialization. If provided, overrides ````generatePreview```` and ````returnByValue```.
    :returns: A tuple with the following items:

        0. **result** - Evaluation result.
        1. **exceptionDetails** - *(Optional)* Exception details.
    """
    params: T_JSON_DICT = dict()
    params["expression"] = expression
    if object_group is not None:
        params["objectGroup"] = object_group
    if include_command_line_api is not None:
        params["includeCommandLineAPI"] = include_command_line_api
    if silent is not None:
        params["silent"] = silent
    if context_id is not None:
        params["contextId"] = context_id.to_json()
    if return_by_value is not None:
        params["returnByValue"] = return_by_value
    if generate_preview is not None:
        params["generatePreview"] = generate_preview
    if user_gesture is not None:
        params["userGesture"] = user_gesture
    if await_promise is not None:
        params["awaitPromise"] = await_promise
    if throw_on_side_effect is not None:
        params["throwOnSideEffect"] = throw_on_side_effect
    if timeout is not None:
        params["timeout"] = timeout.to_json()
    if disable_breaks is not None:
        params["disableBreaks"] = disable_breaks
    if repl_mode is not None:
        params["replMode"] = repl_mode
    if allow_unsafe_eval_blocked_by_csp is not None:
        params["allowUnsafeEvalBlockedByCSP"] = allow_unsafe_eval_blocked_by_csp
    if unique_context_id is not None:
        params["uniqueContextId"] = unique_context_id
    if serialization_options is not None:
        params["serializationOptions"] = serialization_options.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.evaluate",
        "params": params,
    }
    json = yield cmd_dict
    return (
        RemoteObject.from_json(json["result"]),
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def get_isolate_id() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Returns the isolate id.

    **EXPERIMENTAL**

    :returns: The isolate id.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.getIsolateId",
    }
    json = yield cmd_dict
    return str(json["id"])


def get_heap_usage() -> typing.Generator[
    T_JSON_DICT, T_JSON_DICT, typing.Tuple[float, float, float, float]
]:
    """
    Returns the JavaScript heap usage.
    It is the total usage of the corresponding isolate not scoped to a particular Runtime.

    **EXPERIMENTAL**

    :returns: A tuple with the following items:

        0. **usedSize** - Used JavaScript heap size in bytes.
        1. **totalSize** - Allocated JavaScript heap size in bytes.
        2. **embedderHeapUsedSize** - Used size in bytes in the embedder's garbage-collected heap.
        3. **backingStorageSize** - Size in bytes of backing storage for array buffers and external strings.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.getHeapUsage",
    }
    json = yield cmd_dict
    return (
        float(json["usedSize"]),
        float(json["totalSize"]),
        float(json["embedderHeapUsedSize"]),
        float(json["backingStorageSize"]),
    )


def get_properties(
    object_id: RemoteObjectId,
    own_properties: typing.Optional[bool] = None,
    accessor_properties_only: typing.Optional[bool] = None,
    generate_preview: typing.Optional[bool] = None,
    non_indexed_properties_only: typing.Optional[bool] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[
        typing.List[PropertyDescriptor],
        typing.Optional[typing.List[InternalPropertyDescriptor]],
        typing.Optional[typing.List[PrivatePropertyDescriptor]],
        typing.Optional[ExceptionDetails],
    ],
]:
    """
    Returns properties of a given object. Object group of the result is inherited from the target
    object.

    :param object_id: Identifier of the object to return properties for.
    :param own_properties: *(Optional)* If true, returns properties belonging only to the element itself, not to its prototype chain.
    :param accessor_properties_only: **(EXPERIMENTAL)** *(Optional)* If true, returns accessor properties (with getter/setter) only; internal properties are not returned either.
    :param generate_preview: **(EXPERIMENTAL)** *(Optional)* Whether preview should be generated for the results.
    :param non_indexed_properties_only: **(EXPERIMENTAL)** *(Optional)* If true, returns non-indexed properties only.
    :returns: A tuple with the following items:

        0. **result** - Object properties.
        1. **internalProperties** - *(Optional)* Internal object properties (only of the element itself).
        2. **privateProperties** - *(Optional)* Object private properties.
        3. **exceptionDetails** - *(Optional)* Exception details.
    """
    params: T_JSON_DICT = dict()
    params["objectId"] = object_id.to_json()
    if own_properties is not None:
        params["ownProperties"] = own_properties
    if accessor_properties_only is not None:
        params["accessorPropertiesOnly"] = accessor_properties_only
    if generate_preview is not None:
        params["generatePreview"] = generate_preview
    if non_indexed_properties_only is not None:
        params["nonIndexedPropertiesOnly"] = non_indexed_properties_only
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.getProperties",
        "params": params,
    }
    json = yield cmd_dict
    return (
        [PropertyDescriptor.from_json(i) for i in json["result"]],
        [InternalPropertyDescriptor.from_json(i) for i in json["internalProperties"]]
        if json.get("internalProperties", None) is not None
        else None,
        [PrivatePropertyDescriptor.from_json(i) for i in json["privateProperties"]]
        if json.get("privateProperties", None) is not None
        else None,
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def global_lexical_scope_names(
    execution_context_id: typing.Optional[ExecutionContextId] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.List[str]]:
    """
    Returns all let, const and class variables from global scope.

    :param execution_context_id: *(Optional)* Specifies in which execution context to lookup global scope variables.
    :returns:
    """
    params: T_JSON_DICT = dict()
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.globalLexicalScopeNames",
        "params": params,
    }
    json = yield cmd_dict
    return [str(i) for i in json["names"]]


def query_objects(
    prototype_object_id: RemoteObjectId, object_group: typing.Optional[str] = None
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, RemoteObject]:
    """
    :param prototype_object_id: Identifier of the prototype to return objects for.
    :param object_group: *(Optional)* Symbolic group name that can be used to release the results.
    :returns: Array with objects.
    """
    params: T_JSON_DICT = dict()
    params["prototypeObjectId"] = prototype_object_id.to_json()
    if object_group is not None:
        params["objectGroup"] = object_group
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.queryObjects",
        "params": params,
    }
    json = yield cmd_dict
    return RemoteObject.from_json(json["objects"])


def release_object(
    object_id: RemoteObjectId,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Releases remote object with given id.

    :param object_id: Identifier of the object to release.
    """
    params: T_JSON_DICT = dict()
    params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.releaseObject",
        "params": params,
    }
    json = yield cmd_dict


def release_object_group(
    object_group: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Releases all remote objects that belong to a given group.

    :param object_group: Symbolic object group name.
    """
    params: T_JSON_DICT = dict()
    params["objectGroup"] = object_group
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.releaseObjectGroup",
        "params": params,
    }
    json = yield cmd_dict


def run_if_waiting_for_debugger() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Tells inspected instance to run if it was waiting for debugger to attach.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.runIfWaitingForDebugger",
    }
    json = yield cmd_dict


def run_script(
    script_id: ScriptId,
    execution_context_id: typing.Optional[ExecutionContextId] = None,
    object_group: typing.Optional[str] = None,
    silent: typing.Optional[bool] = None,
    include_command_line_api: typing.Optional[bool] = None,
    return_by_value: typing.Optional[bool] = None,
    generate_preview: typing.Optional[bool] = None,
    await_promise: typing.Optional[bool] = None,
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[RemoteObject, typing.Optional[ExceptionDetails]],
]:
    """
    Runs script with given id in a given context.

    :param script_id: Id of the script to run.
    :param execution_context_id: *(Optional)* Specifies in which execution context to perform script run. If the parameter is omitted the evaluation will be performed in the context of the inspected page.
    :param object_group: *(Optional)* Symbolic group name that can be used to release multiple objects.
    :param silent: *(Optional)* In silent mode exceptions thrown during evaluation are not reported and do not pause execution. Overrides ```setPauseOnException```` state.
    :param include_command_line_api: *(Optional)* Determines whether Command Line API should be available during the evaluation.
    :param return_by_value: *(Optional)* Whether the result is expected to be a JSON object which should be sent by value.
    :param generate_preview: *(Optional)* Whether preview should be generated for the result.
    :param await_promise: *(Optional)* Whether execution should ````await``` for resulting value and return once awaited promise is resolved.
    :returns: A tuple with the following items:

        0. **result** - Run result.
        1. **exceptionDetails** - *(Optional)* Exception details.
    """
    params: T_JSON_DICT = dict()
    params["scriptId"] = script_id.to_json()
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    if object_group is not None:
        params["objectGroup"] = object_group
    if silent is not None:
        params["silent"] = silent
    if include_command_line_api is not None:
        params["includeCommandLineAPI"] = include_command_line_api
    if return_by_value is not None:
        params["returnByValue"] = return_by_value
    if generate_preview is not None:
        params["generatePreview"] = generate_preview
    if await_promise is not None:
        params["awaitPromise"] = await_promise
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.runScript",
        "params": params,
    }
    json = yield cmd_dict
    return (
        RemoteObject.from_json(json["result"]),
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None,
    )


def set_async_call_stack_depth(
    max_depth: int,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables or disables async call stacks tracking.

    :param max_depth: Maximum depth of async call stacks. Setting to ```0``` will effectively disable collecting async call stacks (default).
    """
    params: T_JSON_DICT = dict()
    params["maxDepth"] = max_depth
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.setAsyncCallStackDepth",
        "params": params,
    }
    json = yield cmd_dict


def set_custom_object_formatter_enabled(
    enabled: bool,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """


    **EXPERIMENTAL**

    :param enabled:
    """
    params: T_JSON_DICT = dict()
    params["enabled"] = enabled
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.setCustomObjectFormatterEnabled",
        "params": params,
    }
    json = yield cmd_dict


def set_max_call_stack_size_to_capture(
    size: int,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """


    **EXPERIMENTAL**

    :param size:
    """
    params: T_JSON_DICT = dict()
    params["size"] = size
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.setMaxCallStackSizeToCapture",
        "params": params,
    }
    json = yield cmd_dict


def terminate_execution() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Terminate current or next JavaScript execution.
    Will cancel the termination when the outer-most script execution ends.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.terminateExecution",
    }
    json = yield cmd_dict


def add_binding(
    name: str,
    execution_context_id: typing.Optional[ExecutionContextId] = None,
    execution_context_name: typing.Optional[str] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    If executionContextId is empty, adds binding with the given name on the
    global objects of all inspected contexts, including those created later,
    bindings survive reloads.
    Binding function takes exactly one argument, this argument should be string,
    in case of any other input, function throws an exception.
    Each binding function call produces Runtime.bindingCalled notification.

    :param name:
    :param execution_context_id: **(DEPRECATED)** **(EXPERIMENTAL)** *(Optional)* If specified, the binding would only be exposed to the specified execution context. If omitted and ```executionContextName```` is not set, the binding is exposed to all execution contexts of the target. This parameter is mutually exclusive with ````executionContextName````. Deprecated in favor of ````executionContextName```` due to an unclear use case and bugs in implementation (crbug.com/1169639). ````executionContextId```` will be removed in the future.
    :param execution_context_name: *(Optional)* If specified, the binding is exposed to the executionContext with matching name, even for contexts created after the binding is added. See also ````ExecutionContext.name```` and ````worldName```` parameter to ````Page.addScriptToEvaluateOnNewDocument````. This parameter is mutually exclusive with ````executionContextId```.
    """
    params: T_JSON_DICT = dict()
    params["name"] = name
    if execution_context_id is not None:
        params["executionContextId"] = execution_context_id.to_json()
    if execution_context_name is not None:
        params["executionContextName"] = execution_context_name
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.addBinding",
        "params": params,
    }
    json = yield cmd_dict


def remove_binding(name: str) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    This method does not remove binding function from global object but
    unsubscribes current runtime agent from Runtime.bindingCalled notifications.

    :param name:
    """
    params: T_JSON_DICT = dict()
    params["name"] = name
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.removeBinding",
        "params": params,
    }
    json = yield cmd_dict


def get_exception_details(
    error_object_id: RemoteObjectId,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.Optional[ExceptionDetails]]:
    """
    This method tries to lookup and populate exception details for a
    JavaScript Error object.
    Note that the stackTrace portion of the resulting exceptionDetails will
    only be populated if the Runtime domain was enabled at the time when the
    Error was thrown.

    **EXPERIMENTAL**

    :param error_object_id: The error object for which to resolve the exception details.
    :returns:
    """
    params: T_JSON_DICT = dict()
    params["errorObjectId"] = error_object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Runtime.getExceptionDetails",
        "params": params,
    }
    json = yield cmd_dict
    return (
        ExceptionDetails.from_json(json["exceptionDetails"])
        if json.get("exceptionDetails", None) is not None
        else None
    )


@event_class("Runtime.bindingCalled")
@dataclass
class BindingCalled:
    """
    **EXPERIMENTAL**

    Notification is issued every time when binding is called.
    """

    name: str
    payload: str
    #: Identifier of the context where the call was made.
    execution_context_id: ExecutionContextId

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> BindingCalled:
        return cls(
            name=str(json["name"]),
            payload=str(json["payload"]),
            execution_context_id=ExecutionContextId.from_json(
                json["executionContextId"]
            ),
        )


@event_class("Runtime.consoleAPICalled")
@dataclass
class ConsoleAPICalled:
    """
    Issued when console API was called.
    """

    #: Type of the call.
    type_: str
    #: Call arguments.
    args: typing.List[RemoteObject]
    #: Identifier of the context where the call was made.
    execution_context_id: ExecutionContextId
    #: Call timestamp.
    timestamp: Timestamp
    #: Stack trace captured when the call was made. The async stack chain is automatically reported for
    #: the following call types: ``assert``, ``error``, ``trace``, ``warning``. For other types the async call
    #: chain can be retrieved using ``Debugger.getStackTrace`` and ``stackTrace.parentId`` field.
    stack_trace: typing.Optional[StackTrace]
    #: Console context descriptor for calls on non-default console context (not console.*):
    #: 'anonymous#unique-logger-id' for call on unnamed context, 'name#unique-logger-id' for call
    #: on named context.
    context: typing.Optional[str]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ConsoleAPICalled:
        return cls(
            type_=str(json["type"]),
            args=[RemoteObject.from_json(i) for i in json["args"]],
            execution_context_id=ExecutionContextId.from_json(
                json["executionContextId"]
            ),
            timestamp=Timestamp.from_json(json["timestamp"]),
            stack_trace=StackTrace.from_json(json["stackTrace"])
            if json.get("stackTrace", None) is not None
            else None,
            context=str(json["context"])
            if json.get("context", None) is not None
            else None,
        )


@event_class("Runtime.exceptionRevoked")
@dataclass
class ExceptionRevoked:
    """
    Issued when unhandled exception was revoked.
    """

    #: Reason describing why exception was revoked.
    reason: str
    #: The id of revoked exception, as reported in ``exceptionThrown``.
    exception_id: int

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExceptionRevoked:
        return cls(reason=str(json["reason"]), exception_id=int(json["exceptionId"]))


@event_class("Runtime.exceptionThrown")
@dataclass
class ExceptionThrown:
    """
    Issued when exception was thrown and unhandled.
    """

    #: Timestamp of the exception.
    timestamp: Timestamp
    exception_details: ExceptionDetails

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExceptionThrown:
        return cls(
            timestamp=Timestamp.from_json(json["timestamp"]),
            exception_details=ExceptionDetails.from_json(json["exceptionDetails"]),
        )


@event_class("Runtime.executionContextCreated")
@dataclass
class ExecutionContextCreated:
    """
    Issued when new execution context is created.
    """

    #: A newly created execution context.
    context: ExecutionContextDescription

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExecutionContextCreated:
        return cls(context=ExecutionContextDescription.from_json(json["context"]))


@event_class("Runtime.executionContextDestroyed")
@dataclass
class ExecutionContextDestroyed:
    """
    Issued when execution context is destroyed.
    """

    #: Id of the destroyed context
    execution_context_id: ExecutionContextId
    #: Unique Id of the destroyed context
    execution_context_unique_id: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExecutionContextDestroyed:
        return cls(
            execution_context_id=ExecutionContextId.from_json(
                json["executionContextId"]
            ),
            execution_context_unique_id=str(json["executionContextUniqueId"]),
        )


@event_class("Runtime.executionContextsCleared")
@dataclass
class ExecutionContextsCleared:
    """
    Issued when all executionContexts were cleared in browser
    """

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ExecutionContextsCleared:
        return cls()


@event_class("Runtime.inspectRequested")
@dataclass
class InspectRequested:
    """
    Issued when object should be inspected (for example, as a result of inspect() command line API
    call).
    """

    object_: RemoteObject
    hints: dict
    #: Identifier of the context where the call was made.
    execution_context_id: typing.Optional[ExecutionContextId]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> InspectRequested:
        return cls(
            object_=RemoteObject.from_json(json["object"]),
            hints=dict(json["hints"]),
            execution_context_id=ExecutionContextId.from_json(
                json["executionContextId"]
            )
            if json.get("executionContextId", None) is not None
            else None,
        )



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Schema

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


@dataclass
class Domain:
    """
    Description of the protocol domain.
    """

    #: Domain name.
    name: str

    #: Domain version.
    version: str

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        json["version"] = self.version
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Domain:
        return cls(
            name=str(json["name"]),
            version=str(json["version"]),
        )


def get_domains() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.List[Domain]]:
    """
    Returns supported domains.

    :returns: List of supported domains.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Schema.getDomains",
    }
    json = yield cmd_dict
    return [Domain.from_json(i) for i in json["domains"]]



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: FedCm (experimental)

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


class LoginState(enum.Enum):
    """
    Whether this is a sign-up or sign-in action for this account, i.e.
    whether this account has ever been used to sign in to this RP before.
    """

    SIGN_IN = "SignIn"
    SIGN_UP = "SignUp"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> LoginState:
        return cls(json)


class DialogType(enum.Enum):
    """
    The types of FedCM dialogs.
    """

    ACCOUNT_CHOOSER = "AccountChooser"
    AUTO_REAUTHN = "AutoReauthn"
    CONFIRM_IDP_LOGIN = "ConfirmIdpLogin"
    ERROR = "Error"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> DialogType:
        return cls(json)


class DialogButton(enum.Enum):
    """
    The buttons on the FedCM dialog.
    """

    CONFIRM_IDP_LOGIN_CONTINUE = "ConfirmIdpLoginContinue"
    ERROR_GOT_IT = "ErrorGotIt"
    ERROR_MORE_DETAILS = "ErrorMoreDetails"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> DialogButton:
        return cls(json)


class AccountUrlType(enum.Enum):
    """
    The URLs that each account has
    """

    TERMS_OF_SERVICE = "TermsOfService"
    PRIVACY_POLICY = "PrivacyPolicy"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> AccountUrlType:
        return cls(json)


@dataclass
class Account:
    """
    Corresponds to IdentityRequestAccount
    """

    account_id: str

    email: str

    name: str

    given_name: str

    picture_url: str

    idp_config_url: str

    idp_login_url: str

    login_state: LoginState

    #: These two are only set if the loginState is signUp
    terms_of_service_url: typing.Optional[str] = None

    privacy_policy_url: typing.Optional[str] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["accountId"] = self.account_id
        json["email"] = self.email
        json["name"] = self.name
        json["givenName"] = self.given_name
        json["pictureUrl"] = self.picture_url
        json["idpConfigUrl"] = self.idp_config_url
        json["idpLoginUrl"] = self.idp_login_url
        json["loginState"] = self.login_state.to_json()
        if self.terms_of_service_url is not None:
            json["termsOfServiceUrl"] = self.terms_of_service_url
        if self.privacy_policy_url is not None:
            json["privacyPolicyUrl"] = self.privacy_policy_url
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Account:
        return cls(
            account_id=str(json["accountId"]),
            email=str(json["email"]),
            name=str(json["name"]),
            given_name=str(json["givenName"]),
            picture_url=str(json["pictureUrl"]),
            idp_config_url=str(json["idpConfigUrl"]),
            idp_login_url=str(json["idpLoginUrl"]),
            login_state=LoginState.from_json(json["loginState"]),
            terms_of_service_url=str(json["termsOfServiceUrl"])
            if json.get("termsOfServiceUrl", None) is not None
            else None,
            privacy_policy_url=str(json["privacyPolicyUrl"])
            if json.get("privacyPolicyUrl", None) is not None
            else None,
        )


def enable(
    disable_rejection_delay: typing.Optional[bool] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    :param disable_rejection_delay: *(Optional)* Allows callers to disable the promise rejection delay that would normally happen, if this is unimportant to what's being tested. (step 4 of https://fedidcg.github.io/FedCM/#browser-api-rp-sign-in)
    """
    params: T_JSON_DICT = dict()
    if disable_rejection_delay is not None:
        params["disableRejectionDelay"] = disable_rejection_delay
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.enable",
        "params": params,
    }
    json = yield cmd_dict


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.disable",
    }
    json = yield cmd_dict


def select_account(
    dialog_id: str, account_index: int
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    :param dialog_id:
    :param account_index:
    """
    params: T_JSON_DICT = dict()
    params["dialogId"] = dialog_id
    params["accountIndex"] = account_index
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.selectAccount",
        "params": params,
    }
    json = yield cmd_dict


def click_dialog_button(
    dialog_id: str, dialog_button: DialogButton
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    :param dialog_id:
    :param dialog_button:
    """
    params: T_JSON_DICT = dict()
    params["dialogId"] = dialog_id
    params["dialogButton"] = dialog_button.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.clickDialogButton",
        "params": params,
    }
    json = yield cmd_dict


def open_url(
    dialog_id: str, account_index: int, account_url_type: AccountUrlType
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    :param dialog_id:
    :param account_index:
    :param account_url_type:
    """
    params: T_JSON_DICT = dict()
    params["dialogId"] = dialog_id
    params["accountIndex"] = account_index
    params["accountUrlType"] = account_url_type.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.openUrl",
        "params": params,
    }
    json = yield cmd_dict


def dismiss_dialog(
    dialog_id: str, trigger_cooldown: typing.Optional[bool] = None
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    :param dialog_id:
    :param trigger_cooldown: *(Optional)*
    """
    params: T_JSON_DICT = dict()
    params["dialogId"] = dialog_id
    if trigger_cooldown is not None:
        params["triggerCooldown"] = trigger_cooldown
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.dismissDialog",
        "params": params,
    }
    json = yield cmd_dict


def reset_cooldown() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Resets the cooldown time, if any, to allow the next FedCM call to show
    a dialog even if one was recently dismissed by the user.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "FedCm.resetCooldown",
    }
    json = yield cmd_dict


@event_class("FedCm.dialogShown")
@dataclass
class DialogShown:
    dialog_id: str
    dialog_type: DialogType
    accounts: typing.List[Account]
    #: These exist primarily so that the caller can verify the
    #: RP context was used appropriately.
    title: str
    subtitle: typing.Optional[str]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DialogShown:
        return cls(
            dialog_id=str(json["dialogId"]),
            dialog_type=DialogType.from_json(json["dialogType"]),
            accounts=[Account.from_json(i) for i in json["accounts"]],
            title=str(json["title"]),
            subtitle=str(json["subtitle"])
            if json.get("subtitle", None) is not None
            else None,
        )


@event_class("FedCm.dialogClosed")
@dataclass
class DialogClosed:
    """
    Triggered when a dialog is closed, either by user action, JS abort,
    or a command below.
    """

    dialog_id: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DialogClosed:
        return cls(dialog_id=str(json["dialogId"]))



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Database (experimental)

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


class DatabaseId(str):
    """
    Unique identifier of Database object.
    """

    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> DatabaseId:
        return cls(json)

    def __repr__(self):
        return "DatabaseId({})".format(super().__repr__())


@dataclass
class Database:
    """
    Database object.
    """

    #: Database ID.
    id_: DatabaseId

    #: Database domain.
    domain: str

    #: Database name.
    name: str

    #: Database version.
    version: str

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["id"] = self.id_.to_json()
        json["domain"] = self.domain
        json["name"] = self.name
        json["version"] = self.version
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Database:
        return cls(
            id_=DatabaseId.from_json(json["id"]),
            domain=str(json["domain"]),
            name=str(json["name"]),
            version=str(json["version"]),
        )


@dataclass
class Error:
    """
    Database error.
    """

    #: Error message.
    message: str

    #: Error code.
    code: int

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["message"] = self.message
        json["code"] = self.code
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Error:
        return cls(
            message=str(json["message"]),
            code=int(json["code"]),
        )


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disables database tracking, prevents database events from being sent to the client.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Database.disable",
    }
    json = yield cmd_dict


def enable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables database tracking, database events will now be delivered to the client.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Database.enable",
    }
    json = yield cmd_dict


def execute_sql(
    database_id: DatabaseId, query: str
) -> typing.Generator[
    T_JSON_DICT,
    T_JSON_DICT,
    typing.Tuple[
        typing.Optional[typing.List[str]],
        typing.Optional[typing.List[typing.Any]],
        typing.Optional[Error],
    ],
]:
    """
    :param database_id:
    :param query:
    :returns: A tuple with the following items:

        0. **columnNames** -
        1. **values** -
        2. **sqlError** -
    """
    params: T_JSON_DICT = dict()
    params["databaseId"] = database_id.to_json()
    params["query"] = query
    cmd_dict: T_JSON_DICT = {
        "method": "Database.executeSQL",
        "params": params,
    }
    json = yield cmd_dict
    return (
        [str(i) for i in json["columnNames"]]
        if json.get("columnNames", None) is not None
        else None,
        [i for i in json["values"]] if json.get("values", None) is not None else None,
        Error.from_json(json["sqlError"])
        if json.get("sqlError", None) is not None
        else None,
    )


def get_database_table_names(
    database_id: DatabaseId,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.List[str]]:
    """
    :param database_id:
    :returns:
    """
    params: T_JSON_DICT = dict()
    params["databaseId"] = database_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Database.getDatabaseTableNames",
        "params": params,
    }
    json = yield cmd_dict
    return [str(i) for i in json["tableNames"]]


@event_class("Database.addDatabase")
@dataclass
class AddDatabase:
    database: Database

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> AddDatabase:
        return cls(database=Database.from_json(json["database"]))



# coding: utf-8


import asyncio
import logging
import logging.handlers
import random
import time

import mss

logger = logging.getLogger("demo")
logging.basicConfig(level=10)


try:
    import zendriver as uc
except (ModuleNotFoundError, ImportError):
    import os
    import sys

    sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
    import zendriver as uc

_monitor = mss.mss().monitors[0]
SCREEN_WIDTH = _monitor["width"]
NUM_WINS = SCREEN_WIDTH // 325  # as to not fill up the screen to much


class Timing:
    def __init__(self):
        self.start = None
        self.stop = None
        self.taken = None

    def __enter__(self):
        self.start = time.monotonic()
        return self

    def __exit__(self, *args, **kwargs):
        self.stop = time.monotonic()
        self.taken = self.stop - self.start
        print("taken:", self.taken, "seconds")


async def main():
    driver = await uc.start()

    URL1 = "https://www.bet365.com"
    URL2 = "https://www.nowsecure.nl"

    print(
        "the startup speed of the windows depend on the line speed and availability/load on the servers"
    )
    await driver.get(URL2)

    for _ in range(NUM_WINS):
        if _ % 2 == 0:
            await driver.get(URL1, new_window=True)
        else:
            await driver.get(URL2, new_window=True)

    await driver

    grid = await driver.tile_windows(max_columns=NUM_WINS)
    await driver.sleep(5)
    for tab in driver.tabs:
        await tab.maximize()

    for _ in range(15):
        randbox = lambda: random.choice(grid)
        for i, tab in enumerate(driver.tabs):
            await tab.activate()
            await tab.set_window_size(*randbox())
    await driver

    for i, tab in enumerate(driver):
        if i >= len(grid):
            i = len(grid) - i
        await tab.set_window_size(*grid[i])
        await tab.sleep()

    await asyncio.gather(
        *[move_circle(tab, i % 2) for (i, tab) in enumerate(driver.tabs)]
    )

    nowsecure_pages = [tab for tab in driver.tabs if "nowsecure" in tab.url]

    await asyncio.gather(
        *[tab.get("https://nowsecure.nl/mouse.html") for tab in nowsecure_pages]
    )
    await driver.tile_windows(max_columns=NUM_WINS)
    await asyncio.gather(*[mouse_move(tab) for tab in nowsecure_pages])

    b365pages = [tab for tab in driver.tabs if "bet365" in tab.url]
    # await driver.tile_windows()
    await driver.sleep(1)

    positions = await driver.tile_windows(b365pages)
    # await asyncio.gather(*[tab.set_window_size(*pos) for (tab, pos) in zip(b365pages, positions)])
    await asyncio.gather(*[flash_spans(tab, i) for (i, tab) in enumerate(b365pages)])
    await asyncio.gather(*[scroll_task(tab) for tab in b365pages])

    for i, tab in enumerate(driver):
        try:
            await tab.get("https://www.google.com")
            await tab.activate()
            # skip first tab
            if tab == driver.main_tab:
                print("skipping main tab")
                continue
        except:
            pass
        await tab.close()

    for i, tab in enumerate(driver):
        try:
            if i == 0:
                continue
            # await driver.tile_windows(i)
            await tab.activate()
            # skip first tab
            await tab.close()

        except:
            pass
    print("TBCI=", driver.connection.listener.time_before_considered_idle)
    await driver.stop()


async def scroll_task(tab):
    await tab.scroll_up(200)
    spans = await tab.select_all("span")
    [
        await s.scroll_into_view()
        # since scroll_into_view does not return a value
        # we can abuse 'or' to run 2 functions in a list comprehension
        or await s.flash()
        for s in reversed(spans)
    ]
    [
        (await tab.scroll_up(n // 2) or await tab.scroll_down(n))
        # since the above does not return a value
        # we can abuse 'or' to run even more while doing a list comprehension
        or print("tab %s scrolling down : %d" % (tab, n))
        for n in range(0, 75, 15)
    ]


async def mouse_move(tab):
    await tab.activate()
    boxes = await tab.select_all(".box")
    for box in boxes:
        await box.mouse_move()


async def move_circle(tab, x=0):
    window_id, bounds = await tab.get_window()

    old_left, old_top = bounds.left, bounds.top
    old_width, old_height = bounds.width, bounds.height

    center = int(old_left), int(old_top)

    for coord in uc.util.circle(*center, radius=100, num=1050, dir=x):
        new_left, new_top = int(coord[0]), int(coord[1])
        await tab.set_window_size(new_left, new_top)

    for coord in uc.util.circle(*center, radius=250, num=500, dir=x):
        new_left, new_top = int(coord[0] / 2), int(coord[1] / 2)
        # await tab.set_window_size(old_left, old_top)
        await tab.set_window_size(new_left, new_top)

    await tab.set_window_size(old_left, old_top, old_width, old_height)


async def flash_spans(tab, i):
    logger.info("flashing spans. i=%d , tab=%s, url=%s" % (i, tab, tab.url))
    # await tab.fullscreen()
    elems = await tab.select_all("span")
    # await tab.medimize()
    await tab.activate()
    for elem in elems:
        await elem.flash(duration=0.25)
        await elem.scroll_into_view()


if __name__ == "__main__":
    with Timing() as t:
        uc.loop().run_until_complete(main())



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: PerformanceTimeline (experimental)

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT

from . import dom
from . import network
from . import page


@dataclass
class LargestContentfulPaint:
    """
    See https://github.com/WICG/LargestContentfulPaint and largest_contentful_paint.idl
    """

    render_time: network.TimeSinceEpoch

    load_time: network.TimeSinceEpoch

    #: The number of pixels being painted.
    size: float

    #: The id attribute of the element, if available.
    element_id: typing.Optional[str] = None

    #: The URL of the image (may be trimmed).
    url: typing.Optional[str] = None

    node_id: typing.Optional[dom.BackendNodeId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["renderTime"] = self.render_time.to_json()
        json["loadTime"] = self.load_time.to_json()
        json["size"] = self.size
        if self.element_id is not None:
            json["elementId"] = self.element_id
        if self.url is not None:
            json["url"] = self.url
        if self.node_id is not None:
            json["nodeId"] = self.node_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> LargestContentfulPaint:
        return cls(
            render_time=network.TimeSinceEpoch.from_json(json["renderTime"]),
            load_time=network.TimeSinceEpoch.from_json(json["loadTime"]),
            size=float(json["size"]),
            element_id=str(json["elementId"])
            if json.get("elementId", None) is not None
            else None,
            url=str(json["url"]) if json.get("url", None) is not None else None,
            node_id=dom.BackendNodeId.from_json(json["nodeId"])
            if json.get("nodeId", None) is not None
            else None,
        )


@dataclass
class LayoutShiftAttribution:
    previous_rect: dom.Rect

    current_rect: dom.Rect

    node_id: typing.Optional[dom.BackendNodeId] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["previousRect"] = self.previous_rect.to_json()
        json["currentRect"] = self.current_rect.to_json()
        if self.node_id is not None:
            json["nodeId"] = self.node_id.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> LayoutShiftAttribution:
        return cls(
            previous_rect=dom.Rect.from_json(json["previousRect"]),
            current_rect=dom.Rect.from_json(json["currentRect"]),
            node_id=dom.BackendNodeId.from_json(json["nodeId"])
            if json.get("nodeId", None) is not None
            else None,
        )


@dataclass
class LayoutShift:
    """
    See https://wicg.github.io/layout-instability/#sec-layout-shift and layout_shift.idl
    """

    #: Score increment produced by this event.
    value: float

    had_recent_input: bool

    last_input_time: network.TimeSinceEpoch

    sources: typing.List[LayoutShiftAttribution]

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["value"] = self.value
        json["hadRecentInput"] = self.had_recent_input
        json["lastInputTime"] = self.last_input_time.to_json()
        json["sources"] = [i.to_json() for i in self.sources]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> LayoutShift:
        return cls(
            value=float(json["value"]),
            had_recent_input=bool(json["hadRecentInput"]),
            last_input_time=network.TimeSinceEpoch.from_json(json["lastInputTime"]),
            sources=[LayoutShiftAttribution.from_json(i) for i in json["sources"]],
        )


@dataclass
class TimelineEvent:
    #: Identifies the frame that this event is related to. Empty for non-frame targets.
    frame_id: page.FrameId

    #: The event type, as specified in https://w3c.github.io/performance-timeline/#dom-performanceentry-entrytype
    #: This determines which of the optional "details" fields is present.
    type_: str

    #: Name may be empty depending on the type.
    name: str

    #: Time in seconds since Epoch, monotonically increasing within document lifetime.
    time: network.TimeSinceEpoch

    #: Event duration, if applicable.
    duration: typing.Optional[float] = None

    lcp_details: typing.Optional[LargestContentfulPaint] = None

    layout_shift_details: typing.Optional[LayoutShift] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["frameId"] = self.frame_id.to_json()
        json["type"] = self.type_
        json["name"] = self.name
        json["time"] = self.time.to_json()
        if self.duration is not None:
            json["duration"] = self.duration
        if self.lcp_details is not None:
            json["lcpDetails"] = self.lcp_details.to_json()
        if self.layout_shift_details is not None:
            json["layoutShiftDetails"] = self.layout_shift_details.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> TimelineEvent:
        return cls(
            frame_id=page.FrameId.from_json(json["frameId"]),
            type_=str(json["type"]),
            name=str(json["name"]),
            time=network.TimeSinceEpoch.from_json(json["time"]),
            duration=float(json["duration"])
            if json.get("duration", None) is not None
            else None,
            lcp_details=LargestContentfulPaint.from_json(json["lcpDetails"])
            if json.get("lcpDetails", None) is not None
            else None,
            layout_shift_details=LayoutShift.from_json(json["layoutShiftDetails"])
            if json.get("layoutShiftDetails", None) is not None
            else None,
        )


def enable(
    event_types: typing.List[str],
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Previously buffered events would be reported before method returns.
    See also: timelineEventAdded

    :param event_types: The types of event to report, as specified in https://w3c.github.io/performance-timeline/#dom-performanceentry-entrytype The specified filter overrides any previous filters, passing empty filter disables recording. Note that not all types exposed to the web platform are currently supported.
    """
    params: T_JSON_DICT = dict()
    params["eventTypes"] = [i for i in event_types]
    cmd_dict: T_JSON_DICT = {
        "method": "PerformanceTimeline.enable",
        "params": params,
    }
    json = yield cmd_dict


@event_class("PerformanceTimeline.timelineEventAdded")
@dataclass
class TimelineEventAdded:
    """
    Sent when a performance timeline event is added. See reportPerformanceTimeline method.
    """

    event: TimelineEvent

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> TimelineEventAdded:
        return cls(event=TimelineEvent.from_json(json["event"]))



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Cast (experimental)

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


@dataclass
class Sink:
    name: str

    id_: str

    #: Text describing the current session. Present only if there is an active
    #: session on the sink.
    session: typing.Optional[str] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        json["id"] = self.id_
        if self.session is not None:
            json["session"] = self.session
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Sink:
        return cls(
            name=str(json["name"]),
            id_=str(json["id"]),
            session=str(json["session"])
            if json.get("session", None) is not None
            else None,
        )


def enable(
    presentation_url: typing.Optional[str] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Starts observing for sinks that can be used for tab mirroring, and if set,
    sinks compatible with ``presentationUrl`` as well. When sinks are found, a
    ``sinksUpdated`` event is fired.
    Also starts observing for issue messages. When an issue is added or removed,
    an ``issueUpdated`` event is fired.

    :param presentation_url: *(Optional)*
    """
    params: T_JSON_DICT = dict()
    if presentation_url is not None:
        params["presentationUrl"] = presentation_url
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.enable",
        "params": params,
    }
    json = yield cmd_dict


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Stops observing for sinks and issues.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.disable",
    }
    json = yield cmd_dict


def set_sink_to_use(sink_name: str) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets a sink to be used when the web page requests the browser to choose a
    sink via Presentation API, Remote Playback API, or Cast SDK.

    :param sink_name:
    """
    params: T_JSON_DICT = dict()
    params["sinkName"] = sink_name
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.setSinkToUse",
        "params": params,
    }
    json = yield cmd_dict


def start_desktop_mirroring(
    sink_name: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Starts mirroring the desktop to the sink.

    :param sink_name:
    """
    params: T_JSON_DICT = dict()
    params["sinkName"] = sink_name
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.startDesktopMirroring",
        "params": params,
    }
    json = yield cmd_dict


def start_tab_mirroring(
    sink_name: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Starts mirroring the tab to the sink.

    :param sink_name:
    """
    params: T_JSON_DICT = dict()
    params["sinkName"] = sink_name
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.startTabMirroring",
        "params": params,
    }
    json = yield cmd_dict


def stop_casting(sink_name: str) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Stops the active Cast session on the sink.

    :param sink_name:
    """
    params: T_JSON_DICT = dict()
    params["sinkName"] = sink_name
    cmd_dict: T_JSON_DICT = {
        "method": "Cast.stopCasting",
        "params": params,
    }
    json = yield cmd_dict


@event_class("Cast.sinksUpdated")
@dataclass
class SinksUpdated:
    """
    This is fired whenever the list of available sinks changes. A sink is a
    device or a software surface that you can cast to.
    """

    sinks: typing.List[Sink]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SinksUpdated:
        return cls(sinks=[Sink.from_json(i) for i in json["sinks"]])


@event_class("Cast.issueUpdated")
@dataclass
class IssueUpdated:
    """
    This is fired whenever the outstanding issue/error message changes.
    ``issueMessage`` is empty if there is no issue.
    """

    issue_message: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> IssueUpdated:
        return cls(issue_message=str(json["issueMessage"]))



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Console

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


@dataclass
class ConsoleMessage:
    """
    Console message.
    """

    #: Message source.
    source: str

    #: Message severity.
    level: str

    #: Message text.
    text: str

    #: URL of the message origin.
    url: typing.Optional[str] = None

    #: Line number in the resource that generated this message (1-based).
    line: typing.Optional[int] = None

    #: Column number in the resource that generated this message (1-based).
    column: typing.Optional[int] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["source"] = self.source
        json["level"] = self.level
        json["text"] = self.text
        if self.url is not None:
            json["url"] = self.url
        if self.line is not None:
            json["line"] = self.line
        if self.column is not None:
            json["column"] = self.column
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ConsoleMessage:
        return cls(
            source=str(json["source"]),
            level=str(json["level"]),
            text=str(json["text"]),
            url=str(json["url"]) if json.get("url", None) is not None else None,
            line=int(json["line"]) if json.get("line", None) is not None else None,
            column=int(json["column"])
            if json.get("column", None) is not None
            else None,
        )


def clear_messages() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Does nothing.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Console.clearMessages",
    }
    json = yield cmd_dict


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disables console domain, prevents further console messages from being reported to the client.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Console.disable",
    }
    json = yield cmd_dict


def enable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables console domain, sends the messages collected so far to the client by means of the
    ``messageAdded`` notification.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Console.enable",
    }
    json = yield cmd_dict


@event_class("Console.messageAdded")
@dataclass
class MessageAdded:
    """
    Issued when new console message is added.
    """

    #: Console message that has been added.
    message: ConsoleMessage

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> MessageAdded:
        return cls(message=ConsoleMessage.from_json(json["message"]))



import asyncio
import base64
import os

import zendriver as zd


async def main():
    path_image = r"Your Image Path"
    out_dir = r"."
    async with await zd.start() as browser:
        page = browser.main_tab
        await page.get("https://translate.yandex.com/en/ocr")
        await (await page.select('input[type="file"]')).send_file(path_image)
        await asyncio.sleep(3)

        async with page.expect_download() as download_ex:
            await (await page.select("#downloadButton")).mouse_click()
            download = await download_ex.value
        print(download.suggested_filename)
        with open(os.path.join(out_dir, download.suggested_filename), "wb") as fw:
            bytes_file = base64.b64decode(download.url.split(",", 1)[-1])
            fw.write(bytes_file)


if __name__ == "__main__":
    asyncio.run(main())



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: BluetoothEmulation (experimental)

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


class CentralState(enum.Enum):
    """
    Indicates the various states of Central.
    """

    ABSENT = "absent"
    POWERED_OFF = "powered-off"
    POWERED_ON = "powered-on"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> CentralState:
        return cls(json)


class GATTOperationType(enum.Enum):
    """
    Indicates the various types of GATT event.
    """

    CONNECTION = "connection"
    DISCOVERY = "discovery"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> GATTOperationType:
        return cls(json)


class CharacteristicWriteType(enum.Enum):
    """
    Indicates the various types of characteristic write.
    """

    WRITE_DEFAULT_DEPRECATED = "write-default-deprecated"
    WRITE_WITH_RESPONSE = "write-with-response"
    WRITE_WITHOUT_RESPONSE = "write-without-response"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> CharacteristicWriteType:
        return cls(json)


class CharacteristicOperationType(enum.Enum):
    """
    Indicates the various types of characteristic operation.
    """

    READ = "read"
    WRITE = "write"
    SUBSCRIBE_TO_NOTIFICATIONS = "subscribe-to-notifications"
    UNSUBSCRIBE_FROM_NOTIFICATIONS = "unsubscribe-from-notifications"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> CharacteristicOperationType:
        return cls(json)


class DescriptorOperationType(enum.Enum):
    """
    Indicates the various types of descriptor operation.
    """

    READ = "read"
    WRITE = "write"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> DescriptorOperationType:
        return cls(json)


@dataclass
class ManufacturerData:
    """
    Stores the manufacturer data
    """

    #: Company identifier
    #: https://bitbucket.org/bluetooth-SIG/public/src/main/assigned_numbers/company_identifiers/company_identifiers.yaml
    #: https://usb.org/developers
    key: int

    #: Manufacturer-specific data (Encoded as a base64 string when passed over JSON)
    data: str

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["key"] = self.key
        json["data"] = self.data
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ManufacturerData:
        return cls(
            key=int(json["key"]),
            data=str(json["data"]),
        )


@dataclass
class ScanRecord:
    """
    Stores the byte data of the advertisement packet sent by a Bluetooth device.
    """

    name: typing.Optional[str] = None

    uuids: typing.Optional[typing.List[str]] = None

    #: Stores the external appearance description of the device.
    appearance: typing.Optional[int] = None

    #: Stores the transmission power of a broadcasting device.
    tx_power: typing.Optional[int] = None

    #: Key is the company identifier and the value is an array of bytes of
    #: manufacturer specific data.
    manufacturer_data: typing.Optional[typing.List[ManufacturerData]] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        if self.name is not None:
            json["name"] = self.name
        if self.uuids is not None:
            json["uuids"] = [i for i in self.uuids]
        if self.appearance is not None:
            json["appearance"] = self.appearance
        if self.tx_power is not None:
            json["txPower"] = self.tx_power
        if self.manufacturer_data is not None:
            json["manufacturerData"] = [i.to_json() for i in self.manufacturer_data]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ScanRecord:
        return cls(
            name=str(json["name"]) if json.get("name", None) is not None else None,
            uuids=[str(i) for i in json["uuids"]]
            if json.get("uuids", None) is not None
            else None,
            appearance=int(json["appearance"])
            if json.get("appearance", None) is not None
            else None,
            tx_power=int(json["txPower"])
            if json.get("txPower", None) is not None
            else None,
            manufacturer_data=[
                ManufacturerData.from_json(i) for i in json["manufacturerData"]
            ]
            if json.get("manufacturerData", None) is not None
            else None,
        )


@dataclass
class ScanEntry:
    """
    Stores the advertisement packet information that is sent by a Bluetooth device.
    """

    device_address: str

    rssi: int

    scan_record: ScanRecord

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["deviceAddress"] = self.device_address
        json["rssi"] = self.rssi
        json["scanRecord"] = self.scan_record.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ScanEntry:
        return cls(
            device_address=str(json["deviceAddress"]),
            rssi=int(json["rssi"]),
            scan_record=ScanRecord.from_json(json["scanRecord"]),
        )


@dataclass
class CharacteristicProperties:
    """
    Describes the properties of a characteristic. This follows Bluetooth Core
    Specification BT 4.2 Vol 3 Part G 3.3.1. Characteristic Properties.
    """

    broadcast: typing.Optional[bool] = None

    read: typing.Optional[bool] = None

    write_without_response: typing.Optional[bool] = None

    write: typing.Optional[bool] = None

    notify: typing.Optional[bool] = None

    indicate: typing.Optional[bool] = None

    authenticated_signed_writes: typing.Optional[bool] = None

    extended_properties: typing.Optional[bool] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        if self.broadcast is not None:
            json["broadcast"] = self.broadcast
        if self.read is not None:
            json["read"] = self.read
        if self.write_without_response is not None:
            json["writeWithoutResponse"] = self.write_without_response
        if self.write is not None:
            json["write"] = self.write
        if self.notify is not None:
            json["notify"] = self.notify
        if self.indicate is not None:
            json["indicate"] = self.indicate
        if self.authenticated_signed_writes is not None:
            json["authenticatedSignedWrites"] = self.authenticated_signed_writes
        if self.extended_properties is not None:
            json["extendedProperties"] = self.extended_properties
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CharacteristicProperties:
        return cls(
            broadcast=bool(json["broadcast"])
            if json.get("broadcast", None) is not None
            else None,
            read=bool(json["read"]) if json.get("read", None) is not None else None,
            write_without_response=bool(json["writeWithoutResponse"])
            if json.get("writeWithoutResponse", None) is not None
            else None,
            write=bool(json["write"]) if json.get("write", None) is not None else None,
            notify=bool(json["notify"])
            if json.get("notify", None) is not None
            else None,
            indicate=bool(json["indicate"])
            if json.get("indicate", None) is not None
            else None,
            authenticated_signed_writes=bool(json["authenticatedSignedWrites"])
            if json.get("authenticatedSignedWrites", None) is not None
            else None,
            extended_properties=bool(json["extendedProperties"])
            if json.get("extendedProperties", None) is not None
            else None,
        )


def enable(
    state: CentralState, le_supported: bool
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enable the BluetoothEmulation domain.

    :param state: State of the simulated central.
    :param le_supported: If the simulated central supports low-energy.
    """
    params: T_JSON_DICT = dict()
    params["state"] = state.to_json()
    params["leSupported"] = le_supported
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.enable",
        "params": params,
    }
    json = yield cmd_dict


def set_simulated_central_state(
    state: CentralState,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set the state of the simulated central.

    :param state: State of the simulated central.
    """
    params: T_JSON_DICT = dict()
    params["state"] = state.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.setSimulatedCentralState",
        "params": params,
    }
    json = yield cmd_dict


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disable the BluetoothEmulation domain.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.disable",
    }
    json = yield cmd_dict


def simulate_preconnected_peripheral(
    address: str,
    name: str,
    manufacturer_data: typing.List[ManufacturerData],
    known_service_uuids: typing.List[str],
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates a peripheral with ``address``, ``name`` and ``knownServiceUuids``
    that has already been connected to the system.

    :param address:
    :param name:
    :param manufacturer_data:
    :param known_service_uuids:
    """
    params: T_JSON_DICT = dict()
    params["address"] = address
    params["name"] = name
    params["manufacturerData"] = [i.to_json() for i in manufacturer_data]
    params["knownServiceUuids"] = [i for i in known_service_uuids]
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulatePreconnectedPeripheral",
        "params": params,
    }
    json = yield cmd_dict


def simulate_advertisement(
    entry: ScanEntry,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates an advertisement packet described in ``entry`` being received by
    the central.

    :param entry:
    """
    params: T_JSON_DICT = dict()
    params["entry"] = entry.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulateAdvertisement",
        "params": params,
    }
    json = yield cmd_dict


def simulate_gatt_operation_response(
    address: str, type_: GATTOperationType, code: int
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates the response code from the peripheral with ``address`` for a
    GATT operation of ``type``. The ``code`` value follows the HCI Error Codes from
    Bluetooth Core Specification Vol 2 Part D 1.3 List Of Error Codes.

    :param address:
    :param type_:
    :param code:
    """
    params: T_JSON_DICT = dict()
    params["address"] = address
    params["type"] = type_.to_json()
    params["code"] = code
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulateGATTOperationResponse",
        "params": params,
    }
    json = yield cmd_dict


def simulate_characteristic_operation_response(
    characteristic_id: str,
    type_: CharacteristicOperationType,
    code: int,
    data: typing.Optional[str] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates the response from the characteristic with ``characteristicId`` for a
    characteristic operation of ``type``. The ``code`` value follows the Error
    Codes from Bluetooth Core Specification Vol 3 Part F 3.4.1.1 Error Response.
    The ``data`` is expected to exist when simulating a successful read operation
    response.

    :param characteristic_id:
    :param type_:
    :param code:
    :param data: *(Optional)*
    """
    params: T_JSON_DICT = dict()
    params["characteristicId"] = characteristic_id
    params["type"] = type_.to_json()
    params["code"] = code
    if data is not None:
        params["data"] = data
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulateCharacteristicOperationResponse",
        "params": params,
    }
    json = yield cmd_dict


def simulate_descriptor_operation_response(
    descriptor_id: str,
    type_: DescriptorOperationType,
    code: int,
    data: typing.Optional[str] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates the response from the descriptor with ``descriptorId`` for a
    descriptor operation of ``type``. The ``code`` value follows the Error
    Codes from Bluetooth Core Specification Vol 3 Part F 3.4.1.1 Error Response.
    The ``data`` is expected to exist when simulating a successful read operation
    response.

    :param descriptor_id:
    :param type_:
    :param code:
    :param data: *(Optional)*
    """
    params: T_JSON_DICT = dict()
    params["descriptorId"] = descriptor_id
    params["type"] = type_.to_json()
    params["code"] = code
    if data is not None:
        params["data"] = data
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulateDescriptorOperationResponse",
        "params": params,
    }
    json = yield cmd_dict


def add_service(
    address: str, service_uuid: str
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Adds a service with ``serviceUuid`` to the peripheral with ``address``.

    :param address:
    :param service_uuid:
    :returns: An identifier that uniquely represents this service.
    """
    params: T_JSON_DICT = dict()
    params["address"] = address
    params["serviceUuid"] = service_uuid
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.addService",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["serviceId"])


def remove_service(service_id: str) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Removes the service respresented by ``serviceId`` from the simulated central.

    :param service_id:
    """
    params: T_JSON_DICT = dict()
    params["serviceId"] = service_id
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.removeService",
        "params": params,
    }
    json = yield cmd_dict


def add_characteristic(
    service_id: str, characteristic_uuid: str, properties: CharacteristicProperties
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Adds a characteristic with ``characteristicUuid`` and ``properties`` to the
    service represented by ``serviceId``.

    :param service_id:
    :param characteristic_uuid:
    :param properties:
    :returns: An identifier that uniquely represents this characteristic.
    """
    params: T_JSON_DICT = dict()
    params["serviceId"] = service_id
    params["characteristicUuid"] = characteristic_uuid
    params["properties"] = properties.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.addCharacteristic",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["characteristicId"])


def remove_characteristic(
    characteristic_id: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Removes the characteristic respresented by ``characteristicId`` from the
    simulated central.

    :param characteristic_id:
    """
    params: T_JSON_DICT = dict()
    params["characteristicId"] = characteristic_id
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.removeCharacteristic",
        "params": params,
    }
    json = yield cmd_dict


def add_descriptor(
    characteristic_id: str, descriptor_uuid: str
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Adds a descriptor with ``descriptorUuid`` to the characteristic respresented
    by ``characteristicId``.

    :param characteristic_id:
    :param descriptor_uuid:
    :returns: An identifier that uniquely represents this descriptor.
    """
    params: T_JSON_DICT = dict()
    params["characteristicId"] = characteristic_id
    params["descriptorUuid"] = descriptor_uuid
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.addDescriptor",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["descriptorId"])


def remove_descriptor(
    descriptor_id: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Removes the descriptor with ``descriptorId`` from the simulated central.

    :param descriptor_id:
    """
    params: T_JSON_DICT = dict()
    params["descriptorId"] = descriptor_id
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.removeDescriptor",
        "params": params,
    }
    json = yield cmd_dict


def simulate_gatt_disconnection(
    address: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Simulates a GATT disconnection from the peripheral with ``address``.

    :param address:
    """
    params: T_JSON_DICT = dict()
    params["address"] = address
    cmd_dict: T_JSON_DICT = {
        "method": "BluetoothEmulation.simulateGATTDisconnection",
        "params": params,
    }
    json = yield cmd_dict


@event_class("BluetoothEmulation.gattOperationReceived")
@dataclass
class GattOperationReceived:
    """
    Event for when a GATT operation of ``type`` to the peripheral with ``address``
    happened.
    """

    address: str
    type_: GATTOperationType

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> GattOperationReceived:
        return cls(
            address=str(json["address"]),
            type_=GATTOperationType.from_json(json["type"]),
        )


@event_class("BluetoothEmulation.characteristicOperationReceived")
@dataclass
class CharacteristicOperationReceived:
    """
    Event for when a characteristic operation of ``type`` to the characteristic
    respresented by ``characteristicId`` happened. ``data`` and ``writeType`` is
    expected to exist when ``type`` is write.
    """

    characteristic_id: str
    type_: CharacteristicOperationType
    data: typing.Optional[str]
    write_type: typing.Optional[CharacteristicWriteType]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CharacteristicOperationReceived:
        return cls(
            characteristic_id=str(json["characteristicId"]),
            type_=CharacteristicOperationType.from_json(json["type"]),
            data=str(json["data"]) if json.get("data", None) is not None else None,
            write_type=CharacteristicWriteType.from_json(json["writeType"])
            if json.get("writeType", None) is not None
            else None,
        )


@event_class("BluetoothEmulation.descriptorOperationReceived")
@dataclass
class DescriptorOperationReceived:
    """
    Event for when a descriptor operation of ``type`` to the descriptor
    respresented by ``descriptorId`` happened. ``data`` is expected to exist when
    ``type`` is write.
    """

    descriptor_id: str
    type_: CharacteristicOperationType
    data: typing.Optional[str]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DescriptorOperationReceived:
        return cls(
            descriptor_id=str(json["descriptorId"]),
            type_=CharacteristicOperationType.from_json(json["type"]),
            data=str(json["data"]) if json.get("data", None) is not None else None,
        )



import asyncio

import zendriver as zd


async def main() -> None:
    browser = await zd.start()

    # TODO: Read the API response
    page = await browser.get(
        "https://slensky.com/zendriver-examples/api-request.html",
    )

    await browser.stop()


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import zendriver as zd


async def main():
    async with await zd.start() as browser:
        tab = browser.main_tab
        async with tab.expect_request("https://github.com/") as request_info:
            async with tab.expect_response(
                "https://github.githubassets.com/assets/.*"
            ) as response_info:
                await tab.get("https://github.com/")
                await tab.wait_for_ready_state(until="complete")

                req = await request_info.value
                print(req.request_id)

                res = await response_info.value
                print(res.request_id)


if __name__ == "__main__":
    asyncio.run(main())



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Performance

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT


from deprecated.sphinx import deprecated  # type: ignore


@dataclass
class Metric:
    """
    Run-time execution metric.
    """

    #: Metric name.
    name: str

    #: Metric value.
    value: float

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["name"] = self.name
        json["value"] = self.value
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Metric:
        return cls(
            name=str(json["name"]),
            value=float(json["value"]),
        )


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disable collecting and reporting metrics.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Performance.disable",
    }
    json = yield cmd_dict


def enable(
    time_domain: typing.Optional[str] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enable collecting and reporting metrics.

    :param time_domain: *(Optional)* Time domain to use for collecting and reporting duration metrics.
    """
    params: T_JSON_DICT = dict()
    if time_domain is not None:
        params["timeDomain"] = time_domain
    cmd_dict: T_JSON_DICT = {
        "method": "Performance.enable",
        "params": params,
    }
    json = yield cmd_dict


@deprecated(version="1.3")
def set_time_domain(
    time_domain: str,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Sets time domain to use for collecting and reporting duration metrics.
    Note that this must be called before enabling metrics collection. Calling
    this method while metrics collection is enabled returns an error.

    .. deprecated:: 1.3

    **EXPERIMENTAL**

    :param time_domain: Time domain
    """
    params: T_JSON_DICT = dict()
    params["timeDomain"] = time_domain
    cmd_dict: T_JSON_DICT = {
        "method": "Performance.setTimeDomain",
        "params": params,
    }
    json = yield cmd_dict


def get_metrics() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, typing.List[Metric]]:
    """
    Retrieve current values of run-time metrics.

    :returns: Current values for run-time metrics.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Performance.getMetrics",
    }
    json = yield cmd_dict
    return [Metric.from_json(i) for i in json["metrics"]]


@event_class("Performance.metrics")
@dataclass
class Metrics:
    """
    Current values of the metrics.
    """

    #: Current values of the metrics.
    metrics: typing.List[Metric]
    #: Timestamp title.
    title: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Metrics:
        return cls(
            metrics=[Metric.from_json(i) for i in json["metrics"]],
            title=str(json["title"]),
        )



import zendriver as zd
from tests.conftest import CreateBrowser


async def test_multiple_browsers_diff_userdata(create_browser: type[CreateBrowser]):
    config = create_browser().config

    browser1 = await zd.start(config)
    browser2 = await zd.start(config)
    browser3 = await zd.start(config)

    assert not browser1.config.uses_custom_data_dir
    assert not browser2.config.uses_custom_data_dir
    assert not browser3.config.uses_custom_data_dir

    # make sure ports are unique
    ports = {browser1.config.port, browser2.config.port, browser3.config.port}
    assert len(ports) == 3

    # make sure user data dirs are unique
    udds = {
        browser1.config.user_data_dir,
        browser2.config.user_data_dir,
        browser3.config.user_data_dir,
    }
    assert len(udds) == 3

    page1 = await browser1.get("https://example.com/one")
    await page1
    assert page1.target
    assert page1.target.title == "Example Domain"

    page2 = await browser2.get("https://example.com/two")
    await page2
    assert page2.target
    assert page2.target.title == "Example Domain"

    page3 = await browser3.get("https://example.com/three")
    await page3
    assert page3.target
    assert page3.target.title == "Example Domain"

    await browser1.stop()
    await browser2.stop()
    await browser3.stop()



#!/usr/bin/env -S uv run
# /// script
# dependencies = [
#   "requests"
# ]
# ///
"""
Generate and publish a new release of the project.

Intended for use by maintainers only.

This script will:
- Read pyproject.toml to automatically determine the new version number
- Update pyproject.toml with the new version number
- Update zendriver/_version.py with the new version number
- Run `uv sync` to update the lock file
- Update CHANGELOG.md, creating a new section for the release and moving unreleased changes there
- Commit and push changes to pyproject.toml, uv.lock, zendriver/_version.py, and CHANGELOG.md
- Create and push a git tag for the new version
- Create a new release on GitHub ($GITHUB_TOKEN environment variable must be set)

Usage:
    uv run python release.py [--patch | --minor | --major] [--dry-run]
"""

import argparse
import datetime
import os
import re
import subprocess
import sys
from pathlib import Path

# installed via inline script metadata, but mypy doesn't know that
import requests  # type: ignore

REPO_ROOT = Path(__file__).parent.parent

# for creating GitHub release
GITHUB_REPO_SLUG = "stephanlensky/zendriver"

PYPROJECT_TOML = Path("pyproject.toml")
PYPROJECT_VERSION_REGEX = r"^version = \"(?P<version>\d+\.\d+\.\d+)\"$"

VERSION_PY = Path("zendriver/_version.py")

CHANGELOG_MD = Path("CHANGELOG.md")
CHANGELOG_MD_UNRELEASED_REGEX = (
    r"## \[Unreleased\]\s+"
    r"### Fixed\n(?P<fixed>[\s\S]+?)"
    r"### Added\n(?P<added>[\s\S]+?)"
    r"### Changed\n(?P<changed>[\s\S]+?)"
    r"### Removed\n(?P<removed>[\s\S]+?)"
    r"##"
)
CHANGELOG_MD_UNRELEASED_TEMPLATE = (
    "## [Unreleased]\n\n"
    "### Fixed\n\n"
    "### Added\n\n"
    "### Changed\n\n"
    "### Removed\n\n"
)


def set_working_directory() -> None:
    print(f"Changing working directory to {REPO_ROOT}")
    os.chdir(REPO_ROOT)


def ensure_clean_working_directory(message: str | None = None) -> None:
    if subprocess.run(["git", "diff", "--quiet"]).returncode != 0:
        print(
            message
            or "Error: Working directory is not clean, please commit or stash changes"
        )
        sys.exit(1)


def ensure_on_main_branch() -> None:
    current_branch = (
        subprocess.run(["git", "branch", "--show-current"], capture_output=True)
        .stdout.decode()
        .strip()
    )
    if current_branch != "main":
        print("Error: Not on the main branch")
        sys.exit(1)


def ensure_github_token() -> str:
    github_token = os.environ.get("GITHUB_TOKEN")
    if not github_token:
        print("Error: $GITHUB_TOKEN environment variable must be set")
        sys.exit(1)
    return github_token


def get_current_version() -> str:
    content = PYPROJECT_TOML.read_text()
    if match := re.search(PYPROJECT_VERSION_REGEX, content, re.MULTILINE):
        return match.group("version")

    print("Error: Could not find version in pyproject.toml")
    sys.exit(1)


def get_new_version(current_version: str, patch: bool, minor: bool, major: bool) -> str:
    current_major, current_minor, current_patch = map(int, current_version.split("."))
    if major:
        return f"{current_major + 1}.0.0"
    if minor:
        return f"{current_major}.{current_minor + 1}.0"
    if patch:
        return f"{current_major}.{current_minor}.{current_patch + 1}"
    raise ValueError("No version increment specified")


def write_new_version_to_pyproject(new_version: str, dryrun: bool) -> None:
    content = PYPROJECT_TOML.read_text()
    new_content = re.sub(
        PYPROJECT_VERSION_REGEX,
        f'version = "{new_version}"',
        content,
        flags=re.MULTILINE,
    )
    assert content != new_content, "No changes made to pyproject.toml"

    if dryrun:
        print(f"Would update {PYPROJECT_TOML} with new version: {new_version}")
        return
    PYPROJECT_TOML.write_text(new_content)


def write_new_version_to_version_py(new_version: str, dryrun: bool) -> None:
    if dryrun:
        print(f"Would update {VERSION_PY} with new version: {new_version}")
        return

    VERSION_PY.write_text(f'__version__ = "{new_version}"\n')


def sync_lock_file(dryrun: bool) -> None:
    if dryrun:
        print("Would run `uv sync` to update the lock file")
        return

    subprocess.run(["uv", "sync"], check=True)


def write_changelog(new_version: str, dryrun: bool) -> str:
    changelog_content = CHANGELOG_MD.read_text()
    match = re.search(CHANGELOG_MD_UNRELEASED_REGEX, changelog_content)
    if not match:
        print("Error: Could not parse CHANGELOG.md")
        sys.exit(1)
    unreleased_changes = {
        section: match.group(section).strip() for section in match.groupdict()
    }
    if not any(content for content in unreleased_changes.values()):
        print("Error: No changes found in CHANGELOG.md")
        sys.exit(1)

    new_version_changes = "\n\n".join(
        f"### {section.title()}\n\n{unreleased_changes[section]}"
        for section in ("fixed", "added", "changed", "removed")
        if unreleased_changes[section]
    )

    new_version_section = (
        f"## [{new_version}] - {datetime.date.today()}\n\n{new_version_changes}"
    )
    print(f"Changelog:\n{new_version_section}")

    new_changelog_content = re.sub(
        CHANGELOG_MD_UNRELEASED_REGEX,
        f"{CHANGELOG_MD_UNRELEASED_TEMPLATE}{new_version_section}\n\n##",
        changelog_content,
    )

    if dryrun:
        print("Would update CHANGELOG.md with new version section")
    else:
        CHANGELOG_MD.write_text(new_changelog_content)

    return new_version_changes


def show_diff() -> None:
    subprocess.run(["git", "diff", "--color"])


def commit_and_push_changes(new_version: str, dryrun: bool) -> None:
    if dryrun:
        print("Would commit and push changes to the repository")
        return

    files_to_add = [
        "pyproject.toml",
        "uv.lock",
        "CHANGELOG.md",
        "zendriver/_version.py",
    ]
    subprocess.run(["git", "add"] + files_to_add)
    subprocess.run(["git", "commit", "-m", f"Bump version to {new_version}"])
    ensure_clean_working_directory("Error: Working directory is not clean after commit")
    subprocess.run(["git", "push"])


def create_and_push_tag(new_version: str, dryrun: bool) -> None:
    if dryrun:
        print(f"Would create and push tag: v{new_version}")
        return

    subprocess.run(["git", "tag", f"v{new_version}"])
    subprocess.run(["git", "push", "--tags"])


def create_github_release(new_version: str, body: str, dryrun: bool) -> None:
    if dryrun:
        print(f"Would create a new release on GitHub for tag {new_version}")
        return

    github_token = ensure_github_token()

    response = requests.post(
        f"https://api.github.com/repos/{GITHUB_REPO_SLUG}/releases",
        headers={"Authorization": f"token {github_token}"},
        json={"tag_name": f"v{new_version}", "name": f"v{new_version}", "body": body},
    )
    response.raise_for_status()


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Generate and publish a new release of the project."
    )
    parser.add_argument(
        "--patch",
        action="store_true",
        help="Increment the patch version number (e.g. 1.0.0 -> 1.0.1)",
    )
    parser.add_argument(
        "--minor",
        action="store_true",
        help="Increment the minor version number (e.g. 1.0.0 -> 1.1.0)",
    )
    parser.add_argument(
        "--major",
        action="store_true",
        help="Increment the major version number (e.g. 1.0.0 -> 2.0.0)",
    )
    parser.add_argument(
        "--dryrun",
        action="store_true",
        help="Perform a dry run, making no changes to the repository",
    )
    return parser.parse_args()


def main() -> None:
    set_working_directory()
    args = parse_args()
    # prechecks
    if not args.patch and not args.minor and not args.major:
        print("No version increment specified, defaulting to --patch")
        args.patch = True
    elif sum([args.patch, args.minor, args.major]) > 1:
        print("Error: Only one of --patch, --minor, or --major can be specified")
        sys.exit(1)
    if not args.dryrun:
        ensure_clean_working_directory()
        ensure_on_main_branch()
        ensure_github_token()
    if args.dryrun:
        print("Running in dry-run mode, no changes will be made to the repository")

    # calculate new version
    current_version = get_current_version()
    print(f"Detected current version: {current_version}")
    new_version = get_new_version(current_version, args.patch, args.minor, args.major)
    print(f"Calculated new version: {new_version}")

    # create and push the new release
    write_new_version_to_pyproject(new_version, args.dryrun)
    write_new_version_to_version_py(new_version, args.dryrun)
    sync_lock_file(args.dryrun)
    changelog = write_changelog(new_version, args.dryrun)

    show_diff()
    continue_prompt = input("Continue with release? [Y/n]: ")
    if continue_prompt.lower() not in ("y", "yes", ""):
        print("Aborting!")
        sys.exit(1)

    commit_and_push_changes(new_version, args.dryrun)
    create_and_push_tag(new_version, args.dryrun)
    create_github_release(new_version, changelog, args.dryrun)


if __name__ == "__main__":
    main()



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all of the modules.
#
# CDP domain: Profiler

from __future__ import annotations
import enum
import typing
from dataclasses import dataclass
from .util import event_class, T_JSON_DICT

from . import debugger
from . import runtime


@dataclass
class ProfileNode:
    """
    Profile node. Holds callsite information, execution statistics and child nodes.
    """

    #: Unique id of the node.
    id_: int

    #: Function location.
    call_frame: runtime.CallFrame

    #: Number of samples where this node was on top of the call stack.
    hit_count: typing.Optional[int] = None

    #: Child node ids.
    children: typing.Optional[typing.List[int]] = None

    #: The reason of being not optimized. The function may be deoptimized or marked as don't
    #: optimize.
    deopt_reason: typing.Optional[str] = None

    #: An array of source position ticks.
    position_ticks: typing.Optional[typing.List[PositionTickInfo]] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["id"] = self.id_
        json["callFrame"] = self.call_frame.to_json()
        if self.hit_count is not None:
            json["hitCount"] = self.hit_count
        if self.children is not None:
            json["children"] = [i for i in self.children]
        if self.deopt_reason is not None:
            json["deoptReason"] = self.deopt_reason
        if self.position_ticks is not None:
            json["positionTicks"] = [i.to_json() for i in self.position_ticks]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ProfileNode:
        return cls(
            id_=int(json["id"]),
            call_frame=runtime.CallFrame.from_json(json["callFrame"]),
            hit_count=int(json["hitCount"])
            if json.get("hitCount", None) is not None
            else None,
            children=[int(i) for i in json["children"]]
            if json.get("children", None) is not None
            else None,
            deopt_reason=str(json["deoptReason"])
            if json.get("deoptReason", None) is not None
            else None,
            position_ticks=[
                PositionTickInfo.from_json(i) for i in json["positionTicks"]
            ]
            if json.get("positionTicks", None) is not None
            else None,
        )


@dataclass
class Profile:
    """
    Profile.
    """

    #: The list of profile nodes. First item is the root node.
    nodes: typing.List[ProfileNode]

    #: Profiling start timestamp in microseconds.
    start_time: float

    #: Profiling end timestamp in microseconds.
    end_time: float

    #: Ids of samples top nodes.
    samples: typing.Optional[typing.List[int]] = None

    #: Time intervals between adjacent samples in microseconds. The first delta is relative to the
    #: profile startTime.
    time_deltas: typing.Optional[typing.List[int]] = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["nodes"] = [i.to_json() for i in self.nodes]
        json["startTime"] = self.start_time
        json["endTime"] = self.end_time
        if self.samples is not None:
            json["samples"] = [i for i in self.samples]
        if self.time_deltas is not None:
            json["timeDeltas"] = [i for i in self.time_deltas]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Profile:
        return cls(
            nodes=[ProfileNode.from_json(i) for i in json["nodes"]],
            start_time=float(json["startTime"]),
            end_time=float(json["endTime"]),
            samples=[int(i) for i in json["samples"]]
            if json.get("samples", None) is not None
            else None,
            time_deltas=[int(i) for i in json["timeDeltas"]]
            if json.get("timeDeltas", None) is not None
            else None,
        )


@dataclass
class PositionTickInfo:
    """
    Specifies a number of samples attributed to a certain source position.
    """

    #: Source line number (1-based).
    line: int

    #: Number of samples attributed to the source line.
    ticks: int

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["line"] = self.line
        json["ticks"] = self.ticks
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PositionTickInfo:
        return cls(
            line=int(json["line"]),
            ticks=int(json["ticks"]),
        )


@dataclass
class CoverageRange:
    """
    Coverage data for a source range.
    """

    #: JavaScript script source offset for the range start.
    start_offset: int

    #: JavaScript script source offset for the range end.
    end_offset: int

    #: Collected execution count of the source range.
    count: int

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["startOffset"] = self.start_offset
        json["endOffset"] = self.end_offset
        json["count"] = self.count
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CoverageRange:
        return cls(
            start_offset=int(json["startOffset"]),
            end_offset=int(json["endOffset"]),
            count=int(json["count"]),
        )


@dataclass
class FunctionCoverage:
    """
    Coverage data for a JavaScript function.
    """

    #: JavaScript function name.
    function_name: str

    #: Source ranges inside the function with coverage data.
    ranges: typing.List[CoverageRange]

    #: Whether coverage data for this function has block granularity.
    is_block_coverage: bool

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["functionName"] = self.function_name
        json["ranges"] = [i.to_json() for i in self.ranges]
        json["isBlockCoverage"] = self.is_block_coverage
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> FunctionCoverage:
        return cls(
            function_name=str(json["functionName"]),
            ranges=[CoverageRange.from_json(i) for i in json["ranges"]],
            is_block_coverage=bool(json["isBlockCoverage"]),
        )


@dataclass
class ScriptCoverage:
    """
    Coverage data for a JavaScript script.
    """

    #: JavaScript script id.
    script_id: runtime.ScriptId

    #: JavaScript script name or url.
    url: str

    #: Functions contained in the script that has coverage data.
    functions: typing.List[FunctionCoverage]

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = dict()
        json["scriptId"] = self.script_id.to_json()
        json["url"] = self.url
        json["functions"] = [i.to_json() for i in self.functions]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ScriptCoverage:
        return cls(
            script_id=runtime.ScriptId.from_json(json["scriptId"]),
            url=str(json["url"]),
            functions=[FunctionCoverage.from_json(i) for i in json["functions"]],
        )


def disable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.disable",
    }
    json = yield cmd_dict


def enable() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.enable",
    }
    json = yield cmd_dict


def get_best_effort_coverage() -> typing.Generator[
    T_JSON_DICT, T_JSON_DICT, typing.List[ScriptCoverage]
]:
    """
    Collect coverage data for the current isolate. The coverage data may be incomplete due to
    garbage collection.

    :returns: Coverage data for the current isolate.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.getBestEffortCoverage",
    }
    json = yield cmd_dict
    return [ScriptCoverage.from_json(i) for i in json["result"]]


def set_sampling_interval(
    interval: int,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Changes CPU profiler sampling interval. Must be called before CPU profiles recording started.

    :param interval: New sampling interval in microseconds.
    """
    params: T_JSON_DICT = dict()
    params["interval"] = interval
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.setSamplingInterval",
        "params": params,
    }
    json = yield cmd_dict


def start() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.start",
    }
    json = yield cmd_dict


def start_precise_coverage(
    call_count: typing.Optional[bool] = None,
    detailed: typing.Optional[bool] = None,
    allow_triggered_updates: typing.Optional[bool] = None,
) -> typing.Generator[T_JSON_DICT, T_JSON_DICT, float]:
    """
    Enable precise code coverage. Coverage data for JavaScript executed before enabling precise code
    coverage may be incomplete. Enabling prevents running optimized code and resets execution
    counters.

    :param call_count: *(Optional)* Collect accurate call counts beyond simple 'covered' or 'not covered'.
    :param detailed: *(Optional)* Collect block-based coverage.
    :param allow_triggered_updates: *(Optional)* Allow the backend to send updates on its own initiative
    :returns: Monotonically increasing time (in seconds) when the coverage update was taken in the backend.
    """
    params: T_JSON_DICT = dict()
    if call_count is not None:
        params["callCount"] = call_count
    if detailed is not None:
        params["detailed"] = detailed
    if allow_triggered_updates is not None:
        params["allowTriggeredUpdates"] = allow_triggered_updates
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.startPreciseCoverage",
        "params": params,
    }
    json = yield cmd_dict
    return float(json["timestamp"])


def stop() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, Profile]:
    """


    :returns: Recorded profile.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.stop",
    }
    json = yield cmd_dict
    return Profile.from_json(json["profile"])


def stop_precise_coverage() -> typing.Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disable precise code coverage. Disabling releases unnecessary execution count records and allows
    executing optimized code.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.stopPreciseCoverage",
    }
    json = yield cmd_dict


def take_precise_coverage() -> typing.Generator[
    T_JSON_DICT, T_JSON_DICT, typing.Tuple[typing.List[ScriptCoverage], float]
]:
    """
    Collect coverage data for the current isolate, and resets execution counters. Precise code
    coverage needs to have started.

    :returns: A tuple with the following items:

        0. **result** - Coverage data for the current isolate.
        1. **timestamp** - Monotonically increasing time (in seconds) when the coverage update was taken in the backend.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Profiler.takePreciseCoverage",
    }
    json = yield cmd_dict
    return (
        [ScriptCoverage.from_json(i) for i in json["result"]],
        float(json["timestamp"]),
    )


@event_class("Profiler.consoleProfileFinished")
@dataclass
class ConsoleProfileFinished:
    id_: str
    #: Location of console.profileEnd().
    location: debugger.Location
    profile: Profile
    #: Profile title passed as an argument to console.profile().
    title: typing.Optional[str]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ConsoleProfileFinished:
        return cls(
            id_=str(json["id"]),
            location=debugger.Location.from_json(json["location"]),
            profile=Profile.from_json(json["profile"]),
            title=str(json["title"]) if json.get("title", None) is not None else None,
        )


@event_class("Profiler.consoleProfileStarted")
@dataclass
class ConsoleProfileStarted:
    """
    Sent when new profile recording is started using console.profile() call.
    """

    id_: str
    #: Location of console.profile().
    location: debugger.Location
    #: Profile title passed as an argument to console.profile().
    title: typing.Optional[str]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> ConsoleProfileStarted:
        return cls(
            id_=str(json["id"]),
            location=debugger.Location.from_json(json["location"]),
            title=str(json["title"]) if json.get("title", None) is not None else None,
        )


@event_class("Profiler.preciseCoverageDeltaUpdate")
@dataclass
class PreciseCoverageDeltaUpdate:
    """
    **EXPERIMENTAL**

    Reports coverage delta since the last poll (either from an event like this, or from
    ``takePreciseCoverage`` for the current isolate. May only be sent if precise code
    coverage has been started. This event can be trigged by the embedder to, for example,
    trigger collection of coverage data immediately at a certain point in time.
    """

    #: Monotonically increasing time (in seconds) when the coverage update was taken in the backend.
    timestamp: float
    #: Identifier for distinguishing coverage events.
    occasion: str
    #: Coverage data for the current isolate.
    result: typing.List[ScriptCoverage]

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PreciseCoverageDeltaUpdate:
        return cls(
            timestamp=float(json["timestamp"]),
            occasion=str(json["occasion"]),
            result=[ScriptCoverage.from_json(i) for i in json["result"]],
        )



#!/usr/bin/env -S uv run
"""
Copy README.md to docs homepage at docs/index.md.
"""

from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent
README_MD = REPO_ROOT / "README.md"
INDEX_MD = REPO_ROOT / "docs" / "index.md"


def main() -> None:
    readme_content = README_MD.read_text()
    INDEX_MD.write_text(readme_content)
    print("Successfully generated index.md!")


if __name__ == "__main__":
    main()



import asyncio
import re
from typing import Union

from .. import cdp
from .connection import Connection


class BaseRequestExpectation:
    """
    Base class for handling request and response expectations.
    This class provides a context manager to wait for specific network requests and responses
    based on a URL pattern. It sets up handlers for request and response events and provides
    properties to access the request, response, and response body.
    :param tab: The Tab instance to monitor.
    :type tab: Tab
    :param url_pattern: The URL pattern to match requests and responses.
    :type url_pattern: Union[str, re.Pattern[str]]
    """

    def __init__(self, tab: Connection, url_pattern: Union[str, re.Pattern[str]]):
        self.tab = tab
        self.url_pattern = url_pattern
        self.request_future: asyncio.Future[cdp.network.RequestWillBeSent] = (
            asyncio.Future()
        )
        self.response_future: asyncio.Future[cdp.network.ResponseReceived] = (
            asyncio.Future()
        )
        self.request_id: Union[cdp.network.RequestId, None] = None

    async def _request_handler(self, event: cdp.network.RequestWillBeSent):
        """
        Internal handler for request events.
        :param event: The request event.
        :type event: cdp.network.RequestWillBeSent
        """
        if re.fullmatch(self.url_pattern, event.request.url):
            self._remove_request_handler()
            self.request_id = event.request_id
            self.request_future.set_result(event)

    async def _response_handler(self, event: cdp.network.ResponseReceived):
        """
        Internal handler for response events.
        :param event: The response event.
        :type event: cdp.network.ResponseReceived
        """
        if event.request_id == self.request_id:
            self._remove_response_handler()
            self.response_future.set_result(event)

    def _remove_request_handler(self):
        """
        Remove the request event handler.
        """
        self.tab.remove_handlers(cdp.network.RequestWillBeSent, self._request_handler)

    def _remove_response_handler(self):
        """
        Remove the response event handler.
        """
        self.tab.remove_handlers(cdp.network.ResponseReceived, self._response_handler)

    async def __aenter__(self):
        """
        Enter the context manager, adding request and response handlers.
        """
        self.tab.add_handler(cdp.network.RequestWillBeSent, self._request_handler)
        self.tab.add_handler(cdp.network.ResponseReceived, self._response_handler)
        return self

    async def __aexit__(self, *args):
        """
        Exit the context manager, removing request and response handlers.
        """
        self._remove_request_handler()
        self._remove_response_handler()

    @property
    async def request(self):
        """
        Get the matched request.
        :return: The matched request.
        :rtype: cdp.network.Request
        """
        return (await self.request_future).request

    @property
    async def response(self):
        """
        Get the matched response.
        :return: The matched response.
        :rtype: cdp.network.Response
        """
        return (await self.response_future).response

    @property
    async def response_body(self):
        """
        Get the body of the matched response.
        :return: The response body.
        :rtype: str
        """
        request_id = (await self.request_future).request_id
        body = await self.tab.send(cdp.network.get_response_body(request_id=request_id))
        return body


class RequestExpectation(BaseRequestExpectation):
    """
    Class for handling request expectations.
    This class extends `BaseRequestExpectation` and provides a property to access the matched request.
    :param tab: The Tab instance to monitor.
    :type tab: Tab
    :param url_pattern: The URL pattern to match requests.
    :type url_pattern: Union[str, re.Pattern[str]]
    """

    @property
    async def value(self) -> cdp.network.RequestWillBeSent:
        """
        Get the matched request event.
        :return: The matched request event.
        :rtype: cdp.network.RequestWillBeSent
        """
        return await self.request_future


class ResponseExpectation(BaseRequestExpectation):
    """
    Class for handling response expectations.
    This class extends `BaseRequestExpectation` and provides a property to access the matched response.
    :param tab: The Tab instance to monitor.
    :type tab: Tab
    :param url_pattern: The URL pattern to match responses.
    :type url_pattern: Union[str, re.Pattern[str]]
    """

    @property
    async def value(self) -> cdp.network.ResponseReceived:
        """
        Get the matched response event.
        :return: The matched response event.
        :rtype: cdp.network.ResponseReceived
        """
        return await self.response_future


class DownloadExpectation:
    def __init__(self, tab: Connection):
        self.tab = tab
        self.future: asyncio.Future[cdp.browser.DownloadWillBegin] = asyncio.Future()
        # TODO: Improve
        self.default_behavior = (
            self.tab._download_behavior[0] if self.tab._download_behavior else "default"
        )

    async def _handler(self, event: cdp.browser.DownloadWillBegin):
        self._remove_handler()
        self.future.set_result(event)

    def _remove_handler(self):
        self.tab.remove_handlers(cdp.browser.DownloadWillBegin, self._handler)

    async def __aenter__(self):
        """
        Enter the context manager, adding download handler, set download behavior to deny.
        """
        await self.tab.send(
            cdp.browser.set_download_behavior(behavior="deny", events_enabled=True)
        )
        self.tab.add_handler(cdp.browser.DownloadWillBegin, self._handler)
        return self

    async def __aexit__(self, *args):
        """
        Exit the context manager, removing handler, set download behavior to default.
        """
        await self.tab.send(
            cdp.browser.set_download_behavior(behavior=self.default_behavior)
        )
        self._remove_handler()

    @property
    async def value(self) -> cdp.browser.DownloadWillBegin:
        return await self.future



from unittest.mock import Mock

from pytest_mock import MockerFixture

from tests.docs import import_from_path


async def test_infinite_scrolling_tutorial_1(
    mocker: MockerFixture, mock_print: Mock, mock_start: Mock
) -> None:
    module = import_from_path("docs/tutorials/tutorial-code/infinite-scrolling-1.py")

    await module.main()

    mock_print.assert_called_once_with([])


async def test_infinite_scrolling_tutorial_2(
    mocker: MockerFixture, mock_print: Mock, mock_start: Mock
) -> None:
    module = import_from_path("docs/tutorials/tutorial-code/infinite-scrolling-2.py")

    await module.main()

    mock_print.assert_has_calls(
        [mocker.call(f"Card {i}") for i in range(1, 11)],
    )


async def test_infinite_scrolling_tutorial_3(
    mocker: MockerFixture, mock_print: Mock, mock_start: Mock
) -> None:
    module = import_from_path("docs/tutorials/tutorial-code/infinite-scrolling-3.py")

    await module.main()

    mock_print.assert_has_calls(
        [
            mocker.call("Loaded new cards. Current count:", 10),
            mocker.call("Loaded new cards. Current count:", 20),
            mocker.call("Loaded new cards. Current count:", 30),
            mocker.call("Lucky card found: Card 27"),
        ]
    )
